module jp.ku6.windows.game.audio;

import core.thread;
import std.c.stdio;
import jp.ku6.windows.win32api.windef;
import jp.ku6.windows.win32api.winbase;
import jp.ku6.windows.win32api.winuser;
import jp.ku6.windows.win32api.objbase;
import jp.ku6.windows.win32api.wtypes;
import jp.ku6.windows.win32api.mmsystem;
import jp.ku6.windows.directx.dsound;
import org.xiph.libvorbis.codec;
import org.xiph.libvorbis.vorbisfile;
import jp.ku6.windows.game.application;

/++
 + オーディオクラス
 +/
class Audio
{
	/++
	 + 初期化処理
	 +/
	static
	void initialize()
	{
		instance_ = new Audio;
	}

	/++
	 + 終了処理
	 +/
	static
	void terminate()
	{
		delete instance_;
	}

	@property
	{
		/++
		 + Singletonを取得
		 +/
		static
		Audio instance()
		{
			return instance_;
		}
	}

private:

	// singleton
	static
	Audio instance_;

	// DirectSound8 インターフェース
	IDirectSound8 directSound_;

	// プライマリバッファ
	IDirectSoundBuffer primaryBuffer_;

	/+
	 + コンストラクタ
	 +/
	this()
	{
		// DirectSoundインターフェースの取得
		if (FAILED(CoCreateInstance(&CLSID_DirectSound8, null, CLSCTX.CLSCTX_INPROC_SERVER, &IID_IDirectSound8, cast(void**)&directSound_)))
			throw new GameError("Failed get DirectSound");
		scope(failure)
		{
			directSound_.Release();
			directSound_ = null;
		}

		// 初期化
		if (FAILED(directSound_.Initialize(null)))
			throw new GameError("Failed initialize DirectSound");

		// 協調レベルの設定
		if (FAILED(directSound_.SetCooperativeLevel(Application.instance.window.handle, DSSCL_PRIORITY)))
		{
			if (FAILED(directSound_.SetCooperativeLevel(Application.instance.window.handle, DSSCL_NORMAL)))
				throw new GameError("Failed set DirectSound cooperative level");
		}

		// プライマリバッファ作成
		DSBUFFERDESC desc;
		desc.dwSize = DSBUFFERDESC.sizeof;
		desc.dwFlags = DSBCAPS_PRIMARYBUFFER;
		if (FAILED(directSound_.CreateSoundBuffer(&desc, &primaryBuffer_, null)))
			throw new GameError("Failed create DirectSound primary buffer");
		scope(failure)
		{
			primaryBuffer_.Release();
			primaryBuffer_ = null;
		}

		// プライマリバッファフォーマット設定
		WAVEFORMATEX format;
		format.wFormatTag      = WAVE_FORMAT_PCM;
		format.nChannels       = 2;
		format.nSamplesPerSec  = 44100;
		format.wBitsPerSample  = 16;
		format.cbSize          = 0;
		format.nBlockAlign     = cast(ushort)(format.nChannels * (format.wBitsPerSample / 8));
		format.nAvgBytesPerSec = format.nSamplesPerSec * format.nBlockAlign;
		if (FAILED(primaryBuffer_.SetFormat(&format)))
			throw new GameError("Failed set DirectSound primary buffer format");
	}

	/++
	 + デストラクタ
	 +/
	~this()
	{
		// プライマリバッファを開放
		if (primaryBuffer_ !is null)
		{
			primaryBuffer_.Release();
			primaryBuffer_ = null;
		}

		// DirectSoundインターフェースを解放
		if (directSound_ !is null)
		{
			directSound_.Release();
			directSound_ = null;
		}
	}
}

/++
 + 再生する音のインターフェース
 +/
interface Sound
{
	@property
	{
		/// ボリューム -10000 ～ 0
		void volume(int v);
		int volume();
	}

	/++
	 + 再生するファイルをセットする
	 +
	 + Params:
	 +  file = 再生するファイル
	 +/
	void set(SoundSource file);

	/++
	 + 内部で確保したリソースを全てクリアする
	 +/
	void release();

	/++
	 + 再生中か判定する
	 +
	 + Returns:
	 +  再生中の時にtrueを返す
	 +/
	 bool isPlay();

	/++
	 + 再生開始
	 +
	 + Params:
	 +  loop = trueを指定するとループ再生する
	 +  restart = trueを指定すると、再生中でも再度先頭から再生する
	 +  fadeMillisecond = フェードインに使用する時間をミリ秒で指定する
	 +/
	void play(bool loop, bool restart, short fadeMillisecond = 0);

	/++
	 + 停止
	 +
	 + Params:
	 +  fadeMillisecond = フェードアウトに使用する時間をミリ秒で指定する
	 +/
	void stop(short fadeMillisecond = 0);

	/++
	 + 一時停止
	 +/
	void pause();

	/++
	 + 再開
	 +/
	void restart();
}

/++
 + フェードイン/アウトに使用する音量
 +/
struct FadeVolume
{
	/++
	 + フェードイン/アウト
	 +/
	enum Mode
	{
		FADEIN = 1,
		FADEOUT = -1,
	}

	@property
	{
		/++
		 + フェードイン/アウト
		 +/
		Mode mode()
		{
			return mode_;
		}

		/++
		 + ボリュームの値
		 +
		 + Returns:
		 +  最小(0) ～ 最大(0xffff)
		 +/
		int value()
		{
			return value_;
		}
	}

	/++
	 + 音量が有効か調べる
	 +
	 + Returns:
	 +  有効な場合にtrueを返す
	 +/
	bool isEnabled()
	{
		return 0 <= value_ && value_ <= 0xffff;
	}

	/++
	 + 初期化
	 +
	 + Params:
	 +  mode = フェードイン/アウト
	 +  samplesPerSecond = 1秒あたりのデータ数(Hz)
	 +  time = フェードに使用する時間(ミリ秒)
	 +/
	void initialize(Mode mode, uint samplesPerSecond, short time)
	in
	{
		assert(time > 0);
		assert(time < 10000);
	}
	body
	{
		if (time < 0 || 10000 < time)
		{
			value_ = -1;
			return;
		}

		numerator_ = 0;
		denominator_ = samplesPerSecond * time;

		mode_ = mode;

		if (mode_ == Mode.FADEIN)
			value_ = 0;
		else
			value_ = 0xffff;
	}

	/++
	 + 次のデータ用に音量を変更する
	 +/
	void next()
	{
		numerator_ += VOLUME_PER_SAMPLE;
		while (numerator_ > denominator_)
		{
			value_ += mode_;
			numerator_ -= denominator_;
		}
	}

private:

	// 1データ辺りに変動する音量
	immutable static
	uint VOLUME_PER_SAMPLE = 65535000;

	// フェードイン/アウト
	Mode mode_ = Mode.FADEIN;

	// 音量
	int value_ = -1;

	// 音量計算の分母
	uint numerator_ = 1;

	// 音量計算の分子
	uint denominator_ = 1;
}
/++
 + 再生する音ファイルインターフェース
 +/
interface SoundSource
{
	@property
	{
		/// バイトサイズ
		uint length();

		/// 再生時間 (単位:ms)
		int time();

		/// ファイルパス
		string filePath();

		WAVEFORMATEX* format();
	}

	/++
	 + 読み込みカーソルがファイル終端で止まっているか判定
	 +
	 + Returns:
	 +  true = 終端
	 +  false = 終端でない
	 +/
	bool isReadToEnd();

	/++
	 + ファイルオープン
	 +
	 + Returns:
	 +  成功した場合にtrueを返す
	 +/
	bool open();

	/++
	 + ファイルクローズ
	 +
	 + Returns:
	 +  成功した場合にtrueを返す
	 +/
	void close();

	/++
	 + データの読み出し
	 +
	 + Params:
	 +  destination = 読み出したデータを格納するメモリのポインタ
	 +  size = 読み出すデータのサイズ
	 +  loop = ファイル終端まで読み出してまだ足りない場合に先頭から読む場合はtrueを指定する
	 +         falseを指定した場合は、足りない分を0クリアする
	 +/
	void read(ubyte* destination, uint size, bool loop);

	/++
	 + 読み込み位置を変更
	 +
	 + Params:
	 +  offset = 先頭からの位置を指定(単位:ms)
	 +/
	void setReadPosition(uint offset);

	/++
	 + データ読み込み時のフェードインを設定
	 +
	 + Params:
	 +  millisecond = 時間(単位:ms)
	 +/
	void setFadein(short millisecond);

	/++
	 + データ読み込み時のフェードアウトを設定
	 +
	 + Params:
	 +  millisecond = 時間(単位:ms)
	 +/
	void setFadeout(short millisecond);

	/++
	 + フェード処理
	 +
	 + Params:
	 +  buffer = バッファ
	 +  bufferLength = バッファ長(byte)
	 +  bitsPerSample = 1チャンネル1サンプル辺りのbit数
	 +  channels = チャンネル数
	 +  volume = フェード用ボリューム
	 +/
	static
	void fade(void* buffer, uint bufferLength, short bitsPerSample, short channels, FadeVolume* volume)
	{
		if (bitsPerSample == 8)
		{
			uint samples = bufferLength / (channels * 1);
			byte* b = cast(byte*)buffer;
			for (uint i; i < samples; ++i)
			{
				if (!volume.isEnabled())
				{
					if (volume.mode == FadeVolume.Mode.FADEOUT)
					{
						uint length = (samples - i) * channels;
						for (uint j; j < length; ++j)
							b[j] = 0;
					}
					return;
				}

				for (uint j; j < channels; ++j)
				{
					int sample = cast(int)*b;
					sample *= volume.value;
					*b = cast(byte)(sample >> 16);
					++b;
				}

				volume.next();
			}
		}
		else if (bitsPerSample == 16)
		{
			uint samples = bufferLength / (channels * 2);
			short* b = cast(short*)buffer;
			for (uint i; i < samples; ++i)
			{
				if (!volume.isEnabled())
				{
					if (volume.mode == FadeVolume.Mode.FADEOUT)
					{
						uint length = (samples - i) * channels;
						for (uint j; j < length; ++j)
							b[j] = 0;
					}
					return;
				}

				for (uint j; j < channels; ++j)
				{
					int sample = cast(int)*b;
					sample *= volume.value;
					*b = cast(short)(sample >> 16);
					++b;
				}

				volume.next();
			}
		}
	}
}

/++
 + オンメモリ再生クラス
 +
 + 最初にメモリに全てのデータを読み出す
 + ディスクアクセスはなくなるが、メモリを消費する
 + 主にSE用
 +/
class StaticSound : Sound, WindowMessageListener
{
	/++
	 + コンストラクタ
	 +/
	this()
	{
		// Audioの初期化を確認
		if (Audio.instance is null)
			throw new GameError("Audio is not initialized");

		Application.instance.addWindowMessageListener(this);
	}

	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  source = 再生する音データ
	 +/
	this(SoundSource source)
	{
		this();
		set(source);
	}

	/++
	 + デストラクタ
	 +/
	~this()
	{
		Application.instance.removeWindowMessageListener(this);

		// DirectSoundBufferの解放
		if (soundBuffer_ !is null)
		{
			soundBuffer_.Release();
			soundBuffer_ = null;
		}

		if (file_ !is null)
			delete file_;
	}

	@property
	{
		void volume(int v)
		{
			if (v < -10000)
				v = -10000;
			else if(v > 0)
				v = 0;

			if (soundBuffer_ is null)
				return;

			// ステータス確認
			uint status;
			if (soundBuffer_.GetStatus(&status) != DS_OK)
				return;

			// バッファがロストしていないかチェックし、必要ならレストアする
			if ((status & DSBSTATUS_BUFFERLOST) != 0)
			{
				if (soundBuffer_.Restore() != DS_OK)
					return;
			}

			soundBuffer_.SetVolume(v);

			volume_ = v;
		}

		int volume()
		{
			return volume_;
		}
	}

	void set(SoundSource file)
	{
		if (file is null || !file.open())
			return;
		scope(exit) file.close();

		// 既にセット済みならば差し替える
		if (file_ !is null && file_ !is file)
			delete file_;
		file_ = file;

		loop_ = false;
		
		// DirectSoundBufferの解放
		if (soundBuffer_ !is null)
		{
			soundBuffer_.Release();
			soundBuffer_ = null;
		}

		// DirectSoundBufferパラメータセット
		DSBUFFERDESC desc;
		desc.dwSize = DSBUFFERDESC.sizeof;
		desc.dwFlags = DSBCAPS_LOCSOFTWARE | DSBCAPS_CTRLVOLUME | DSBCAPS_STICKYFOCUS;
		desc.dwBufferBytes = file_.length;
		desc.lpwfxFormat = file_.format;

		// DirectSoundBuffer作成
		IDirectSoundBuffer tmpBuffer;
		if (Audio.instance is null || Audio.instance.directSound_ is null || Audio.instance.directSound_.CreateSoundBuffer(&desc, &tmpBuffer, null) != DS_OK)
			return;

		// DirectSoundBuffer8取得
		if (FAILED(tmpBuffer.QueryInterface(&IID_IDirectSoundBuffer8, cast(void**)&soundBuffer_)))
		{
			tmpBuffer.Release();
			soundBuffer_ = null;
			return;
		}

		// DirectSoundBufferはもういらないので解放
		tmpBuffer.Release();

		// サウンドバッファをロック
		LPVOID buffer;
		uint length;
		soundBuffer_.Lock(0, 0, &buffer, &length, null, null, DSBLOCK_ENTIREBUFFER);

		// PCMデータをロード
		file_.read(cast(ubyte*)buffer, length, false);

		// ロック解除
		soundBuffer_.Unlock(buffer, length, null, 0);

		// ファイルをクローズ
		file_.close();
	}

	void release()
	{
		loop_ = false;
		volume_ = 0;
		
		// DirectSoundBufferの解放
		if (soundBuffer_ !is null)
		{
			soundBuffer_.Release();
			soundBuffer_ = null;
		}
	}

	bool isPlay()
	{
		if (soundBuffer_ is null)
			return false;

		// ステータス確認
		uint status;
		if (soundBuffer_.GetStatus(&status) != DS_OK)
			return false;

		// バッファがロストしていないかチェックし、必要ならレストアする
		if ((status & DSBSTATUS_BUFFERLOST) != 0)
		{
			if (soundBuffer_.Restore() != DS_OK)
				return false;
		}

		// 再生フラグを返す
		return (status & DSBSTATUS_PLAYING) != 0;
	}

	void play(bool loop, bool restart, short fadeMillisecond)
	{
		if (soundBuffer_ is null)
			return;

		// ステータス確認
		uint status;
		if (soundBuffer_.GetStatus(&status) != DS_OK)
			return;

		// バッファがロストしていないかチェックし、必要ならレストアする
		if ((status & DSBSTATUS_BUFFERLOST) != 0)
		{
			if (soundBuffer_.Restore() != DS_OK)
				return;
		}

		// 再生カーソルを先頭に持ってくる
		if (restart || (status & DSBSTATUS_PLAYING) == 0)
			soundBuffer_.SetCurrentPosition(0);

		// ループフラグ更新
		loop_ = loop;

		// バッファの再生開始（すでに再生しているときはループフラグだけ変更になる）
		if (loop_)
			soundBuffer_.Play(0, 0, DSBPLAY_LOOPING);
		else
			soundBuffer_.Play(0, 0, 0);
	}

	void stop(short fadeMillisecond)
	{
		if (soundBuffer_ is null)
			return;

		// ステータス確認
		uint status;
		if (soundBuffer_.GetStatus(&status) != DS_OK)
			return;

		// バッファがロストしていないかチェックし、必要ならレストアする
		if ((status & DSBSTATUS_BUFFERLOST) != 0)
		{
			if (soundBuffer_.Restore() != DS_OK)
				return;
		}

		// 再生中ならば停止する
		if ((status & DSBSTATUS_PLAYING) != 0)
			soundBuffer_.Stop();

		// 再生カーソルを先頭へ（stop()後のrestart()に対応）
		soundBuffer_.SetCurrentPosition(0);
	}

	void pause()
	{
		if (soundBuffer_ is null)
			return;

		// ステータス確認
		uint status;
		if (soundBuffer_.GetStatus(&status) != DS_OK)
			return;

		// バッファがロストしていないかチェックし、必要ならレストアする
		if ((status & DSBSTATUS_BUFFERLOST) != 0)
		{
			if (soundBuffer_.Restore() != DS_OK)
				return;
		}

		if ((status & DSBSTATUS_PLAYING) != 0)
			soundBuffer_.Stop();
	}

	void restart()
	{
		if (soundBuffer_ is null)
			return;

		// ステータス確認
		uint status;
		if (soundBuffer_.GetStatus(&status) != DS_OK)
			return;

		// バッファがロストしていないかチェックし、必要ならレストアする
		if ((status & DSBSTATUS_BUFFERLOST) != 0)
		{
			if (soundBuffer_.Restore() != DS_OK)
				return;
		}

		if ((status & DSBSTATUS_PLAYING) == 0)
		{
			if (loop_)
				soundBuffer_.Play(0, 0, DSBPLAY_LOOPING);
			else
				soundBuffer_.Play(0, 0, 0);
		}
	}

	bool procedure(HWND windowHandle, UINT message, WPARAM wParam, LPARAM lParam)
	{
		if (soundBuffer_ is null)
			return true;

		bool active = false;
		switch (message) 
		{
		// ウインドウのアクティブ状態の変更時の処理
		case WM_ACTIVATE:
			if (LOWORD(wParam) != WA_INACTIVE)
				active = true;	
			else
				active = false;
			break;

		// ウインドウのサイズが変更されたときの処理
		case WM_SIZE:
			// 最小化を捕捉する
			if (wParam == SIZE_MINIMIZED)
				active = false;
			else if (wParam == SIZE_RESTORED || wParam == SIZE_MAXIMIZED)
				active = true;
			break;

		// それ以外は何もしない
		default:
			return true;
		}

		// ステータス確認
		uint status;
		if (soundBuffer_.GetStatus(&status) != DS_OK)
			return true;

		// バッファがロストしていないかチェックし、必要ならレストアする
		if ((status & DSBSTATUS_BUFFERLOST) != 0)
		{
			if (soundBuffer_.Restore() != DS_OK)
				return true;
		}

		if (active)
		{
			if (resume_ && (status & DSBSTATUS_PLAYING) == 0)
			{
				if (loop_)
					soundBuffer_.Play(0, 0, DSBPLAY_LOOPING);
				else
					soundBuffer_.Play(0, 0, 0);
			}
		}
		else
		{
			if ((status & DSBSTATUS_PLAYING) != 0)
			{
				soundBuffer_.Stop();
				resume_ = true;
			}
			else
				resume_ = false;
		}

		return true;
	}

private:

	// ループ再生フラグ
	bool loop_;

	// ウインドウ操作で再生再開フラグ
	bool resume_;

	// ボリューム
	int volume_;

	// サウンドバッファ
	IDirectSoundBuffer8 soundBuffer_;

	// 音声ファイル
	SoundSource file_;
}

/++
 + ストリーミング再生クラス
 +
 + ファイルを少しずつ読み出しながら再生するため、省メモリで動作する
 + 主にBGM用
 +/
class StreamSound : Sound, WindowMessageListener
{
	/++
	 + コンストラクタ
	 +/
	this()
	{
		// Audioの初期化を確認
		if (Audio.instance is null)
			throw new GameError("Audio is not initialized");

		// イベントを作成
		notifyEvent_ = CreateEvent(null, FALSE, FALSE, null);

		// スレッドの作成
		streamingThread_ = new Thread(&streaming);
		streamingThread_.start();

		Application.instance.addWindowMessageListener(this);
	}

	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  source = 再生する音データ
	 +/
	this(SoundSource source)
	{
		this();
		set(source);
	}

	/++
	 + デストラクタ
	 +/
	~this()
	{
		Application.instance.removeWindowMessageListener(this);

		synchronized (this)
		{
			// スレッドを生存フラグをOFF
			theradValid_ = false;

			// イベントをシグナル状態にして、Waitを止める
			SetEvent(notifyEvent_);
		}

		// スレッド終了を待機する
		streamingThread_.join();

		// イベントハンドルを閉じる
		if (notifyEvent_ !is null)
		{
			CloseHandle(notifyEvent_);
			notifyEvent_ = null;
		}

		// DirectSoundNotifyを解放
		if (soundNotify_ !is null)
		{
			soundNotify_.Release();
			soundNotify_ = null;
		}

		//	DirectSoundBufferの解放
		if (soundBuffer_ !is null)
		{
			soundBuffer_.Stop();
			soundBuffer_.Release();
			soundBuffer_ = null;
		}

		// ファイルを削除
		if (file_ !is null)
		{
			delete file_;
			file_ = null;
		}
	}

	@property
	{
		void volume(int v)
		{
			if (v < -10000)
				v = -10000;
			else if(v > 0)
				v = 0;

			if (soundBuffer_ is null)
				return;

			synchronized (this)
			{
				uint status;
				if (soundBuffer_.GetStatus(&status) != DS_OK)
					return;

				if ((status & DSBSTATUS_BUFFERLOST) != 0)
				{
					if (soundBuffer_.Restore() != DS_OK)
						return;
				}

				soundBuffer_.SetVolume(v);
			}

			volume_ = volume;
		}

		int volume()
		{
			return volume_;
		}
	}

	void set(SoundSource file)
	{
		if (file is null || !file.open())
			return;

		synchronized (this)
		{
			// 既にサウンドバッファがあるか
			if (soundBuffer_ !is null)
			{
				// 既にセットされているフォーマットと、新しいフォーマットが一致するか
				if (*(file_.format) == *(file.format))
				{
					uint status;
					if (soundBuffer_.GetStatus(&status) == DS_OK)
					{
						if ((status & DSBSTATUS_BUFFERLOST) == 0 || soundBuffer_.Restore() == DS_OK)
						{
							if ((status & DSBSTATUS_PLAYING) != 0)
								soundBuffer_.Stop();

							// ファイルを差し替え
							if (file_ !is null && file_ !is file)
								delete file_;
							file_ = file;

							return;
						}
					}
				}

				// 再作成するデータをクリア

				// DirectSoundNotifyを解放
				if (soundNotify_ !is null)
				{
					soundNotify_.Release();
					soundNotify_ = null;
				}

				//	DirectSoundBufferの解放
				if (soundBuffer_ !is null)
				{
					soundBuffer_.Stop();
					soundBuffer_.Release();
					soundBuffer_ = null;
				}
			}

			// ファイルを差し替え
			if (file_ !is null)
				delete file_;
			file_ = file;

			// 値の初期化
			loop_ = false;
			play_ = false;

			// DirectSoundBufferパラメータセット
			DSBUFFERDESC desc;
			desc.dwSize = DSBUFFERDESC.sizeof;
			desc.dwFlags = DSBCAPS_LOCSOFTWARE | DSBCAPS_CTRLVOLUME | DSBCAPS_STICKYFOCUS | DSBCAPS_CTRLPOSITIONNOTIFY;
			desc.dwBufferBytes = BUFFER_BLOCK_SIZE * BUFFER_BLOCK_COUNT;
			desc.lpwfxFormat = file_.format;

			// DirectSoundBuffer作成
			IDirectSoundBuffer tmpBuffer;
			if (Audio.instance is null || Audio.instance.directSound_ is null || Audio.instance.directSound_.CreateSoundBuffer(&desc, &tmpBuffer, null) != DS_OK)
			{
				file_.close();
				return;
			}

			// DirectSoundBuffer8取得
			if (FAILED(tmpBuffer.QueryInterface(&IID_IDirectSoundBuffer8, cast(void**)&soundBuffer_)))
			{
				file_.close();
				tmpBuffer.Release();
				soundBuffer_ = null;
				return;
			}

			// DirectSoundBufferはもういらないので解放
			tmpBuffer.Release();

			// DirectSoundNotify作成
			if (FAILED(soundBuffer_.QueryInterface(&IID_IDirectSoundNotify8, cast(LPVOID*)&soundNotify_)))
			{
				file_.close();
				soundBuffer_.Release();
				soundBuffer_ = null;
				soundNotify_ = null;
				return;
			}

			//	イベントの発生ポイントの数分作成
			DSBPOSITIONNOTIFY[BUFFER_BLOCK_COUNT] positionNotify;
			for (uint i; i < BUFFER_BLOCK_COUNT; ++i)
			{
				positionNotify[i].dwOffset = BUFFER_BLOCK_SIZE * i;
			    positionNotify[i].hEventNotify = notifyEvent_;
			}

		 	// イベント発生位置を設定
			soundNotify_.SetNotificationPositions(BUFFER_BLOCK_COUNT, positionNotify.ptr);
		}
	}

	void release()
	{
		synchronized (this)
		{
			// DirectSoundNotifyを解放
			if (soundNotify_ !is null)
			{
				soundNotify_.Release();
				soundNotify_ = null;
			}

			//	DirectSoundBufferの解放
			if (soundBuffer_ !is null)
			{
				soundBuffer_.Stop();
				soundBuffer_.Release();
				soundBuffer_ = null;
			}

			// ファイルをクローズ
			file_.close();

			// 値の初期化
			loop_ = false;
			play_ = false;
		}
	}

	bool isPlay()
	{
		return play_;
	}

	void play(bool loop, bool restart, short fadeMillisecond)
	{
		if (soundBuffer_ is null)
			return;

		synchronized (this)
		{
			uint status;
			if (soundBuffer_.GetStatus(&status) != DS_OK)
				return;

			if ((status & DSBSTATUS_BUFFERLOST) != 0)
			{
				if (soundBuffer_.Restore() != DS_OK)
					return;
			}

			if (restart || (status & DSBSTATUS_PLAYING) == 0)
			{
				loop_ = loop;

				// 先頭から再生
				file_.setReadPosition(0);
				if (fadeMillisecond > 0)
					file_.setFadein(fadeMillisecond);
				soundBuffer_.SetCurrentPosition(0);

				// 最初にちょっと読んでおく
				readBlockFromFile(0, 2);

				// バッファの再生開始
				soundBuffer_.Play(0, 0, DSBPLAY_LOOPING);

				play_ = true;
			}
			else
			{
				loop_ = loop;

				// こいつが来るのはかなり稀なタイミングだが・・・
				if (file_.isReadToEnd())
				{
					// 先頭から再生
					file_.setReadPosition(0);
					if (fadeMillisecond > 0)
						file_.setFadein(fadeMillisecond);
					soundBuffer_.SetCurrentPosition(0);

					// 最初にちょっと読んでおく
					readBlockFromFile(0, 2);

					play_ = true;
				}
			}
		}
	}

	void stop(short fadeMillisecond)
	{
		if (soundBuffer_ is null)
			return;

		if (fadeMillisecond > 0)
		{
			file_.setFadeout(fadeMillisecond);
			return;
		}

		synchronized (this)
		{
			uint status;
			if (soundBuffer_.GetStatus(&status) != DS_OK)
				return;

			if ((status & DSBSTATUS_BUFFERLOST) != 0)
			{
				if (soundBuffer_.Restore() != DS_OK)
					return;
			}

			if ((status & DSBSTATUS_PLAYING) != 0)
				soundBuffer_.Stop();

			play_ = false;
		}
	}

	void pause()
	{
		stop(0);
	}

	void restart()
	{
		if (soundBuffer_ is null)
			return;

		synchronized (this)
		{
			uint status;
			if (soundBuffer_.GetStatus(&status) != DS_OK)
				return;

			if ((status & DSBSTATUS_BUFFERLOST) != 0)
			{
				if (soundBuffer_.Restore() != DS_OK)
					return;
			}

			if ((status & DSBSTATUS_PLAYING) == 0)
				soundBuffer_.Play(0, 0, DSBPLAY_LOOPING);
		}
	}

	bool procedure(HWND windowHandle, UINT message, WPARAM wParam, LPARAM lParam)
	{
		if (!play_ || soundBuffer_ is null)
			return true;

		bool active = false;
		switch (message) 
		{
		// ウインドウのアクティブ状態の変更時の処理
		case WM_ACTIVATE:
			if (LOWORD(wParam) != WA_INACTIVE)
				active = true;	
			else
				active = false;
			break;

		// ウインドウのサイズが変更されたときの処理
		case WM_SIZE:
			// 最小化を捕捉する
			if (wParam == SIZE_MINIMIZED)
				active = false;
			else if (wParam == SIZE_RESTORED || wParam == SIZE_MAXIMIZED)
				active = true;
			break;

		// それ以外は何もしない
		default:
			return true;
		}

		synchronized (this)
		{
			uint status;
			if (soundBuffer_.GetStatus(&status) != DS_OK)
				return true;

			if ((status & DSBSTATUS_BUFFERLOST) != 0)
			{
				if (soundBuffer_.Restore() != DS_OK)
					return true;
			}

			if (active)
			{
				if ((status & DSBSTATUS_PLAYING) == 0)
					soundBuffer_.Play(0, 0, DSBPLAY_LOOPING);
			}
			else
			{
				if ((status & DSBSTATUS_PLAYING) != 0)
					soundBuffer_.Stop();
			}
		}

		return true;
	}

private:

	// ストリーミング時に一度にデータを読み出すブロックサイズ
	immutable uint BUFFER_BLOCK_SIZE = 64 * 1024;

	// ストリーミング時に、バッファを分割する個数
	immutable uint BUFFER_BLOCK_COUNT = 4;

	// ストリーミングスレッド継続フラグ
	bool theradValid_;

	// 再生フラグ
	bool play_;

	// ループフラグ
	bool loop_;

	// ボリューム
	int volume_;

	// バッファ充填用イベント
	HANDLE notifyEvent_;

	// DirectSound buffer
	IDirectSoundBuffer8 soundBuffer_;

	// DirectSound notify
	IDirectSoundNotify8 soundNotify_;

	// ファイル読み出しスレッド
	Thread streamingThread_;

	// 読み出すファイル
	SoundSource file_;

	/+
	 + ファイル読み出し処理
	 +
	 + 別スレッドでファイル読み出し処理を行う
	 +/
	void streaming()
	{
		theradValid_ = true;

		uint safetyRead;
		for (;;)
		{
			// DirectSoundBufferからのイベント発生を待つ
			WaitForSingleObject(notifyEvent_, INFINITE);

			synchronized (this)
			{
				if (!theradValid_)
					return;

				if (soundBuffer_ is null)
					continue;

				// ステータス確認
				uint status;
				if (soundBuffer_.GetStatus(&status) != DS_OK)
					continue;

				// バッファがロストしていないかチェックし、必要ならレストアする
				if ((status & DSBSTATUS_BUFFERLOST) != 0)
				{
					if (soundBuffer_.Restore() != DS_OK)
						continue;
				}

				// 現在再生中のブロックを取得
				uint currentBufferBlock;
				soundBuffer_.GetCurrentPosition(null, &currentBufferBlock);
				currentBufferBlock /= BUFFER_BLOCK_SIZE;
				currentBufferBlock = (currentBufferBlock + 2) & 3;

				// 再生中のブロックの次の次のブロックに読み込む
				readBlockFromFile(currentBufferBlock, 1);

				// ループ再生でないときEOFまでよんだか？
				if (file_.isReadToEnd())
				{
					// 安全圏まで再生しておく
					if (++safetyRead >= BUFFER_BLOCK_COUNT)
					{
						if ((status & DSBSTATUS_PLAYING) != 0)
							soundBuffer_.Stop();
					}

					safetyRead = 0;
					ResetEvent(notifyEvent_);
				}
			}
		}
	}

	void readBlockFromFile(uint blockStart, uint blockCount)
	{
		// サウンドバッファをロック
		LPVOID buffer;
		uint length;
		soundBuffer_.Lock(blockStart * BUFFER_BLOCK_SIZE, blockCount * BUFFER_BLOCK_SIZE, &buffer, &length, null, null, 0);

		// PCMデータをロード
		file_.read(cast(ubyte*)buffer, length, loop_);

		// アンロック
		soundBuffer_.Unlock(buffer, length, null, 0);
	}
}

/++
 + .wavファイル
 +/
class WaveFile : SoundSource
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  filePath = ファイルパス
	 +  offset = ファイル先頭からのオフセット。1ファイルに複数のファイルをまとめた時に使用
	 +/
	this(string filePath, uint offset)
	{
		filePath_ = filePath ~ 0;
		fileOffset_ = offset;
	}

	/++
	 + デストラクタ
	 +/
	~this()
	{
		close();
	}

	@property
	{
		uint length()
		{
			return length_;
		}

		int time()
		{
			if (format_.nAvgBytesPerSec == 0)
				return 0;
			else
				return length_ * 1000 / format_.nAvgBytesPerSec;
		}

		string filePath()
		{
			return filePath_[0 .. $ - 1];
		}

		WAVEFORMATEX* format()
		{
			return &format_;
		}
	}

	bool isReadToEnd()
	{
		return readToEnd_;
	}

	bool open()
	{
		// ファイルをオープン
		FILE* file = std.c.stdio.fopen(filePath_.ptr, "rb\0".ptr);	
		if (file is null)
			return false;

		// Waveファイル先頭位置まで移動
		if (fileOffset_ != 0)
			std.c.stdio.fseek(file, fileOffset_, std.c.stdio.SEEK_SET);
		uint offset = fileOffset_;


		uint data;

		// RIFFファイル?
		std.c.stdio.fread(&data, 4, 1, file);
		offset += 4;
		if (data != mmioFOURCC!('R', 'I', 'F', 'F'))
		{
			fclose(file);
			return false;
		}

		// ファイルサイズはスキップ
		std.c.stdio.fseek(file, 4, std.c.stdio.SEEK_CUR);
		offset += 4;

		// Wave形式?
		std.c.stdio.fread(&data, 4, 1, file);
		offset += 4;
		if (data != mmioFOURCC!('W', 'A', 'V', 'E'))
		{
			fclose(file);
			return false;
		}

		// "fmt "の文字列
		std.c.stdio.fread(&data, 4, 1, file);
		offset += 4;
		if (data != mmioFOURCC!('f', 'm', 't', ' '))
		{
			fclose(file);
			return false;
		}

		// fmtチャンクのバイト数
		std.c.stdio.fread(&data, 4, 1, file);
		offset += 4;

		// WaveFormat読み込み
		std.c.stdio.fread(&format_, 16, 1, file);
		offset += 16;
		
		// 使用しないデータをスキップ
		std.c.stdio.fseek(file, data - 16, std.c.stdio.SEEK_CUR);
		offset += data - 16;

		// Dataチャンク検索
		for (;;)
		{
			//	なんのチャンク？
			std.c.stdio.fread(&data, 4, 1, file);
			offset += 4;

			// EOFチェック
			if (std.c.stdio.feof(file))
			{
				fclose(file);
				return false;
			}
			
			// チャンク名を確認
			if (data == mmioFOURCC!('d','a','t','a'))
				break;
			else
			{
				// 違ったらそのチャンクをスキップする
				std.c.stdio.fread(&data, 4, 1, file);
				offset += 4;

				std.c.stdio.fseek(file, data, std.c.stdio.SEEK_CUR);
				offset += data;
			}
		}
		
		// 波形データサイズ
		std.c.stdio.fread(&data, 4, 1, file);
		offset += 4;

		file_ = file;
		filePath_ = filePath;
		readToEnd_ = false;
		readCursor_ = 0;
		dataOffset_ = offset;
		length_ = data;

		return true;
	}

	void close()
	{
		if (file_ !is null)
		{
			std.c.stdio.fclose(file_);
			file_ = null;
		}

		readToEnd_ = true;
		readCursor_ = length_;
	}

	void read(ubyte* destination, uint size, bool loop)
	{
		// もう読めなければゼロクリアする
		if (readToEnd_ || file_ is null)
		{
			destination[0 .. size] = 0;
			return;
		}

		for (uint readSize; readSize < size; )
		{
			if (size - readSize > length_ - readCursor_)
			{
				// そのまま読み込むとEOFに到達する場合

				// とりあえず読めるだけは読む
				std.c.stdio.fread(destination + readSize, 1, length_ - readCursor_, file_);
				readSize += length_ - readCursor_;

				if (loop)	
				{
					// 繰り返し再生の場合は先頭に戻る
					readCursor_ = 0;
					std.c.stdio.fseek(file_, dataOffset_, std.c.stdio.SEEK_SET);
				}
				else
				{
					// 繰り返しでない場合はゼロクリアする
					destination[readSize .. size] = 0;

					readCursor_ = 0;
					readToEnd_ = true;
					readSize = size;
					break;
				}
			}
			else
			{
				// そのまま読んで問題ない場合

				// 読み込み
				std.c.stdio.fread(destination + readSize, 1, size - readSize, file_);

				// 読み込みカーソル更新
				readCursor_ += size - readSize;

				// 念のため
				if (readCursor_ >= length_)
				{
					readCursor_ -= length_;
					if (!loop)
						readToEnd_ = true;
				}

				break;
			}
		}

		if (fadeVolume_.isEnabled())
		{
			SoundSource.fade(destination, size, format_.wBitsPerSample, format_.nChannels, &fadeVolume_);
			if (!fadeVolume_.isEnabled())
			{
				if (fadeVolume_.mode == FadeVolume.Mode.FADEOUT)
					readToEnd_ = true; // ループであろうがなかろうが終端フラグを立てる
			}
		}
	}

	void setReadPosition(uint offset)
	{
		readCursor_ = format_.nAvgBytesPerSec * offset / 1000;

		uint atomicByte = format_.nChannels * format.wBitsPerSample / 8;
		if (readCursor_ != 0)
			readCursor_ -= readCursor_ % atomicByte;

		if (readCursor_ >= length_)
		{
			readCursor_ = length_;
			readToEnd_ = true;
		}
		else
		{
			std.c.stdio.fseek(file_, dataOffset_ + readCursor_, std.c.stdio.SEEK_SET);
			readToEnd_ = false;
		}
	}

	void setFadein(short millisecond)
	{
		fadeVolume_.initialize(FadeVolume.Mode.FADEIN, format_.nSamplesPerSec, millisecond);
	}

	void setFadeout(short millisecond)
	{
		fadeVolume_.initialize(FadeVolume.Mode.FADEOUT, format_.nSamplesPerSec, millisecond);
	}

private:

	// データを終端まで読み込んだフラグ
	bool readToEnd_ = true;

	// ファイル先頭からWaveファイルまでのオフセット(Byte)
	uint fileOffset_;

	// ファイル先頭から波形データまでのオフセット(Byte)
	uint dataOffset_;

	// 波形データのバイト数
	uint length_;

	// 読み出し位置
	uint readCursor_;

	// ファイルポインタ
	FILE* file_;

	// ファイル名
	string filePath_;

	// Waveデータのフォーマット
	WAVEFORMATEX format_;

	// フェードイン/アウト用音量
	FadeVolume fadeVolume_;
}

class OggVorbisFile : SoundSource
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  filePath = ファイルパス
	 +  offset = ファイル先頭からのオフセット。1ファイルに複数のファイルをまとめた時に使用
	 +/
	this(string filePath, uint offset)
	{
		filePath_ = filePath ~ 0;
		fileOffset_ = offset;
	}

	/++
	 + デストラクタ
	 +/
	~this()
	{
		close();
	}

	@property
	{
		uint length()
		{
			return length_;
		}

		int time()
		{
			return cast(int)(ov_time_total(&oggVorbisFile_, -1) * 1000);
		}

		string filePath()
		{
			return filePath_[0 .. $ - 1];
		}

		WAVEFORMATEX* format()
		{
			return &format_;
		}
	}

	bool isReadToEnd()
	{
		return readToEnd_;
	}

	bool open()
	{
		// ファイル自体をオープン
		FILE* file = std.c.stdio.fopen(filePath_.ptr, "rb\0".ptr);
		if (file is null)
			return false;

		// アーカイブファイル内の個別ファイル位置まで移動
		std.c.stdio.fseek(file, fileOffset_, std.c.stdio.SEEK_SET);

		// OggVorbisファイルとしてオープン
		if (ov_open(file, &oggVorbisFile_, null, 0) != 0)
			return false;

		// Seek可能か確認
		if (ov_seekable(&oggVorbisFile_) == 0)
		{
			close();
			return false;
		}

		// WAVEFORMATEX構造体の設定
		const(vorbis_info)* vorbisInfo = ov_info(&oggVorbisFile_, -1);
		format_.nChannels = cast(ushort)(vorbisInfo.channels & 0x0000FFFF);
		format_.wFormatTag = WAVE_FORMAT_PCM;
		format_.nSamplesPerSec = vorbisInfo.rate;
		format_.wBitsPerSample = 16;
		format_.nBlockAlign = cast(ushort)(format_.nChannels * (format_.wBitsPerSample / 8));
		format_.nAvgBytesPerSec = format_.nSamplesPerSec * format_.nBlockAlign;

		//	デコードしたときのPCMデータのバイト数を算出
		length_ = cast(uint)(ov_pcm_total(&oggVorbisFile_, -1)) * format_.nBlockAlign;

		readToEnd_ = false;

		return true;
	}

	void close()
	{
		ov_clear(&oggVorbisFile_);

		readToEnd_ = true;
	}

	void read(ubyte* destination, uint size, bool loop)
	{
		if (readToEnd_)
		{
			destination[0 .. size] = 0;
			return;
		}

		for (uint read; read < size;)
		{
			// 読み込み
			int r =  ov_read(&oggVorbisFile_, cast(char*)(destination + read),  size - read, 0, 2, 1, null);

			// デーコードされたサイズを更新
			read += r;

			// ファイル終端をチェック
			if (r == 0)
			{
				if (loop)
					setReadPosition(0);
				else
				{
					readToEnd_ = true;
					destination[read .. size] = 0;
					break;
				}
			}
		}

		if (fadeVolume_.isEnabled())
		{
			SoundSource.fade(destination, size, format_.wBitsPerSample, format_.nChannels, &fadeVolume_);
			if (!fadeVolume_.isEnabled())
			{
				if (fadeVolume_.mode == FadeVolume.Mode.FADEOUT)
					readToEnd_ = true; // ループであろうがなかろうが終端フラグを立てる
			}
		}
	}

	void setReadPosition(uint offset)
	{
		int result = ov_time_seek(&oggVorbisFile_, cast(double)offset / 1000);

		readToEnd_ = false;
	}

	void setFadein(short millisecond)
	{
		fadeVolume_.initialize(FadeVolume.Mode.FADEIN, format_.nSamplesPerSec, millisecond);
	}

	void setFadeout(short millisecond)
	{
		fadeVolume_.initialize(FadeVolume.Mode.FADEOUT, format_.nSamplesPerSec, millisecond);
	}

private:

	// データを終端まで読み込んだフラグ
	bool readToEnd_ = true;

	// ファイル先頭からWaveファイルまでのオフセット(Byte)
	uint fileOffset_;

	// 波形データのバイト数
	uint length_;

	// ファイル名
	string filePath_;

	// Waveデータのフォーマット
	WAVEFORMATEX format_;

	// OggVorbis_File
	OggVorbis_File oggVorbisFile_;

	// フェードイン/アウト用音量
	FadeVolume fadeVolume_;
}
