module jp.ku6.windows.game.graphics;

import std.c.stdio;
import std.c.stdlib;
import std.stream;
import jp.ku6.windows.win32api.windows;
import jp.ku6.windows.directx.dxgiformat;
import jp.ku6.windows.directx.dxgitype;
import jp.ku6.windows.directx.dxgi;
import jp.ku6.windows.directx.d3dcommon;
import jp.ku6.windows.directx.d3d11;
import jp.ku6.windows.directx.d3dcompiler;
import jp.ku6.windows.game.application;

/++
 + グラフィックスクラス
 +/
class Graphics : WindowMessageListener
{
	/++
	 + 初期化処理
	 +
	 + 描画領域は最初に設定したら後は変えられません
	 +
	 + Params:
	 +  width = 描画領域の幅
	 +  height = 描画領域の高さ
	 +/
	static
	void initialize(int width, int height)
	{
		instance_ = new Graphics(width, height);
	}

	/++
	 + 終了処理
	 +/
	static
	void terminate()
	{
		delete instance_;
	}

	@property
	{
		/++
		 + Singletonを取得
		 +/
		static
		Graphics instance()
		{
			return instance_;
		}

		/++
		 + 描画領域の幅を取得
		 +
		 + ウィンドウのクライアント領域やバックバッファのサイズではなく、
		 + 一時的にレンダリングする仮想バックバッファのサイズ
		 + ゲーム中では画面が常にこのサイズとしてプログラムを作れば良い
		 +
		 + Returns:
		 +  描画領域の幅
		 +/
		int width()
		{
			return width_;
		}

		/++
		 + 描画領域の高さを取得
		 +
		 + ウィンドウのクライアント領域やバックバッファのサイズではなく、
		 + 一時的にレンダリングする仮想バックバッファのサイズ
		 + ゲーム中では画面が常にこのサイズとしてプログラムを作れば良い
		 +
		 + Returns:
		 +  描画領域の高さ
		 +/
		int height()
		{
			return height_;
		}

		/++
		 + ID3D11Deviceを直接取得する
		 +
		 + Returns:
		 +  ID3D11Device
		 +/
		ID3D11Device device()
		{
			return device_;
		}

		/++
		 + ID3D11DeviceContextを直接取得する
		 +
		 + マルチスレッドは対応していないので取得出来るのはこのコンテキストだけです
		 +
		 + Returns:
		 +  ID3D11DeviceContext
		 +/
		ID3D11DeviceContext immediateContext()
		{
			return immediateContext_;
		}

		/++
		 + シェーダコンパイル関数を取得
		 +
		 + 要るのはこれくらい？
		 + 他にも要るようならばDLLのハンドルを返すようにする
		 +
		 + Returns:
		 +  シェーダコンパイル関数
		 +/
		D3DCompileFunction D3DCompile()
		{
			return D3DCompile_;
		}
	}

	/++
	 + 描画開始
	 +/
	void begin()
	{
		// レンダリング先が描画用一時テクスチャになるので、描画領域のサイズを指定
		spriteRenderer_.setScreenSize(width_, height_);

		// ビューポートの設定
		viewPort_.Width = width_;
		viewPort_.Height = height_;
		immediateContext_.RSSetViewports(1, &viewPort_);

		// 背景をクリア
		float[4] clearColor = [0, 0, 0, 1.0f];
		immediateContext_.ClearRenderTargetView(textureRenderTargetView_, clearColor.ptr);

		// レンダリング対象のテクスチャをセット
		immediateContext_.OMSetRenderTargets(1, &textureRenderTargetView_, null);
	}

	/++
	 + 描画終了
	 +/
	void end()
	{
		int clientWidth = Application.instance.window.clientWidth;
		int clientHeight = Application.instance.window.clientHeight;

		// レンダリング先がバックバッファになるので、バックバッファサイズ(ウィンドウクライアント領域サイズ)の設定
		spriteRenderer_.setScreenSize(clientWidth, clientHeight);

		// ビューポートの設定
		viewPort_.Width = clientWidth;
		viewPort_.Height = clientHeight;
		immediateContext_.RSSetViewports(1, &viewPort_);


		// 背景をクリア
		float[4] clearColor = [0, 0, 0, 0];
		immediateContext_.ClearRenderTargetView(backBufferRenerTargetView_, clearColor.ptr);

		// レンダリングターゲットをバックバッファに設定
		immediateContext_.OMSetRenderTargets(1, &backBufferRenerTargetView_, null);


		// 描画対象テクスチャスプライトとしてバックバッファに描画する
		SpriteRenderer.Constants* constants = &spriteRenderer_.constants_;
		constants.centerX = 0;
		constants.centerY = 0;
		constants.inclineX = 0;
		constants.inclineY = 0;
		constants.u = 0;
		constants.v = 0;
		constants.width = width_;
		constants.height = height_;
		constants.textureWidth = renderTargetWidth_;
		constants.textureHeight = renderTargetHeight_;
		constants.rotation = 0;
		constants.color = 0xffffffff;

		if (clientWidth * height_ > clientHeight * width_)
		{
			// 縦の拡大率の方が小さい時
			constants.scaleX = constants.scaleY = cast(float)clientHeight / cast(float)height_;

			// 縦の拡大率に合わせるので横が余る
			constants.x = cast(float)(clientWidth - ((width_ * clientHeight) / height_)) * 0.5f;
			constants.y = 0;
		}
		else
		{
			// 横の拡大率の方が小さい時
			constants.scaleX = constants.scaleY = cast(float)clientWidth / cast(float)width_;

			// 横の拡大率に合わせるので縦が余る
			constants.x = 0;
			constants.y = cast(float)(clientHeight - ((height_ * clientWidth) / width_)) * 0.5f;
		}

		// 描画処理
		spriteRenderer_.render(immediateContext_, textureResourceView_);

		swapChain_.Present(0, 0);


		// 描画対象テクスチャを Pixel shader から外す(次のOMSetのため)
		ID3D11ShaderResourceView nullResourceView = null;
		immediateContext_.PSSetShaderResources(0, 1, &nullResourceView);
	}

	override
	bool procedure(HWND windowHandle, UINT message, WPARAM wParam, LPARAM lParam)
	{
		switch (message) 
		{
		// ウインドウのサイズが変更されたときの処理
		case WM_SIZE:
			if (swapChain_ !is null)
			{
				if (wParam == SIZE_RESTORED || wParam == SIZE_MAXIMIZED)
				{
					ID3D11RenderTargetView nullRenderTargetView = null;
					immediateContext_.OMSetRenderTargets(1, &nullRenderTargetView, null);

					backBufferRenerTargetView_.Release();

					swapChain_.ResizeBuffers(2, 0, 0, DXGI_FORMAT.DXGI_FORMAT_R8G8B8A8_UNORM, 0);

					// バックバッファを2Dテクスチャとして取得
					ID3D11Texture2D backBuffer;
					swapChain_.GetBuffer(0, &IID_ID3D11Texture2D, cast(void**)&backBuffer);
					scope(exit) backBuffer.Release();

					// バックバッファをレンダリングターゲットとするインターフェースを作成
					if (FAILED(device_.CreateRenderTargetView(backBuffer, null, &backBufferRenerTargetView_)))
						throw new GameError("Failed create render target view");
				}
			}
			break;

		default:
			break;
		}

		return true;
	}

private:

	// singleton
	static
	Graphics instance_;

	// d3d11.dllのハンドル
	HMODULE d3d11Dll_;

	// d3dcompiler_xx.dllのハンドル
	HMODULE d3dcompiler_;

	// シェーダコンパイル関数
	D3DCompileFunction D3DCompile_;

	// Swap chain
	IDXGISwapChain swapChain_;

	// Device
	ID3D11Device device_;

	// Device context
	ID3D11DeviceContext immediateContext_;

	// バックバッファ
	ID3D11RenderTargetView backBufferRenerTargetView_;

	// ゲーム画面レンダリング先テクスチャ
	ID3D11Texture2D texture_;

	// ゲーム画面レンダリング先
	ID3D11RenderTargetView textureRenderTargetView_;

	// ゲーム画面レンダリング先
	ID3D11ShaderResourceView textureResourceView_;

	// View port
	D3D11_VIEWPORT viewPort_;

	// Sprite描画ステータス
	SpriteRenderer spriteRenderer_;

	// 描画領域の幅
	int width_;

	// 描画領域の高さ
	int height_;

	// 描画対象テクスチャの幅
	int renderTargetWidth_;

	// 描画対象テクスチャの高さ
	int renderTargetHeight_;
	
	/+
	 + コンストラクタ
	 +/
	this(int width, int height)
	{
		width_ = width;
		height_ = height;

		//--------------------------------------------------------------
		// 使用するDLLの読み込み

		// Direct3D 11
		d3d11Dll_ = LoadLibrary("d3d11.dll\0".ptr);
		if (d3d11Dll_ is null)
			throw new GameError("Failed load d3d11.dll");

		// shader compiler
		d3dcompiler_ = LoadLibrary(D3DCOMPILER_DLL.ptr);
		if (d3dcompiler_ is null)
		{
			release();
			throw new GameError("Failed load d3dcompiler.dll");
		}

		//--------------------------------------------------------------
		// 基本となるDirect 3D Deviceとスワップチェインを用意

		// D3D11CreateDeviceAndSwapChainのアドレスを取得
		PFN_D3D11_CREATE_DEVICE_AND_SWAP_CHAIN D3D11CreateDeviceAndSwapChain = cast(PFN_D3D11_CREATE_DEVICE_AND_SWAP_CHAIN)GetProcAddress(d3d11Dll_, "D3D11CreateDeviceAndSwapChain\0".ptr);
		if (D3D11CreateDeviceAndSwapChain is null)
		{
			release();
			throw new GameError("Failed get D3D11CreateDeviceAndSwapChain function");
		}

		// Swap chainの設定
		DXGI_SWAP_CHAIN_DESC swapChainDesc;
		swapChainDesc.BufferDesc.Width = Application.instance.window.clientWidth;
		swapChainDesc.BufferDesc.Height = Application.instance.window.clientHeight;
		swapChainDesc.BufferDesc.Format = DXGI_FORMAT.DXGI_FORMAT_R8G8B8A8_UNORM;
		swapChainDesc.BufferDesc.RefreshRate.Numerator = 60;
		swapChainDesc.BufferDesc.RefreshRate.Denominator = 1;
		swapChainDesc.SampleDesc.Count = 1;
		swapChainDesc.SampleDesc.Quality = 0;
		swapChainDesc.BufferUsage = DXGI_USAGE_RENDER_TARGET_OUTPUT;
		swapChainDesc.OutputWindow = Application.instance.window.handle;
		swapChainDesc.BufferCount = 1;
		swapChainDesc.Windowed = TRUE;

		// IDXGISwapChain, ID3D11Device, ID3D11DeviceContextを作成
		D3D_FEATURE_LEVEL[2] featureLevels;
		featureLevels[0] = D3D_FEATURE_LEVEL.D3D_FEATURE_LEVEL_11_0;
		featureLevels[1] = D3D_FEATURE_LEVEL.D3D_FEATURE_LEVEL_10_1;
		if (FAILED(D3D11CreateDeviceAndSwapChain(null, D3D_DRIVER_TYPE.D3D_DRIVER_TYPE_HARDWARE, null, 0, featureLevels.ptr, featureLevels.length, D3D11_SDK_VERSION, &swapChainDesc, &swapChain_, &device_, null, &immediateContext_)))
		{
			release();
			throw new GameError("Failed create device");
		}

		// バックバッファを2Dテクスチャとして取得
		ID3D11Texture2D backBuffer;
		swapChain_.GetBuffer(0, &IID_ID3D11Texture2D, cast(void**)&backBuffer);

		// バックバッファをレンダリングターゲットとするインターフェースを作成
		if (FAILED(device_.CreateRenderTargetView(backBuffer, null, &backBufferRenerTargetView_)))
		{
			backBuffer.Release();
			release();
			throw new GameError("Failed create render target view");
		}

		backBuffer.Release();


		//--------------------------------------------------------------
		// Viewportの設定

		viewPort_.MinDepth = 0.0f;
		viewPort_.MaxDepth = 1.0f;
		viewPort_.TopLeftX = 0;
		viewPort_.TopLeftY = 0;


		//--------------------------------------------------------------
		// 描画対象テクスチャの作成

		renderTargetWidth_ = width;
		renderTargetHeight_ = height;
	
		int size;

		size = 1;
		for (;;)
		{
			if (size >= renderTargetWidth_)
			{
				renderTargetWidth_ = size;
				break;
			}
			size *= 2;
		}

		size = 1;
		for (;;)
		{
			if (size >= renderTargetHeight_)
			{
				renderTargetHeight_ = size;
				break;
			}
			size *= 2;
		}

		D3D11_TEXTURE2D_DESC textureDesc;
		textureDesc.Width = renderTargetWidth_;
		textureDesc.Height = renderTargetHeight_;
		textureDesc.MipLevels = 1;
		textureDesc.ArraySize = 1;
		textureDesc.Format = DXGI_FORMAT.DXGI_FORMAT_R8G8B8A8_UNORM;
		textureDesc.SampleDesc.Count = 1;
		textureDesc.SampleDesc.Quality = 0;
		textureDesc.Usage = D3D11_USAGE.D3D11_USAGE_DEFAULT;
		textureDesc.BindFlags = D3D11_BIND_FLAG.D3D11_BIND_SHADER_RESOURCE | D3D11_BIND_FLAG.D3D11_BIND_RENDER_TARGET;
		textureDesc.CPUAccessFlags = 0;

		if (FAILED(device_.CreateTexture2D(&textureDesc, null, &texture_)))
		{
			release();
			throw new GameError("Failed create render target texture");
		}

		if (FAILED(device_.CreateShaderResourceView(texture_, null, &textureResourceView_)))
		{
			release();
			throw new GameError("Failed create render target texture resource view");
		}

		if (FAILED(device_.CreateRenderTargetView(texture_, null, &textureRenderTargetView_)))
		{
			release();
			throw new GameError("Failed create render target texture render target view");
		}


		//--------------------------------------------------------------
		// シェーダコンパイラの取得

		// D3DCompileのアドレスを取得
		D3DCompile_ = cast(D3DCompileFunction)GetProcAddress(d3dcompiler_, "D3DCompile\0".ptr);
		if (D3DCompile_ is null)
		{
			release();
			throw new GameError("Failed get D3DCompile function");
		}


		//--------------------------------------------------------------
		// Spriteの描画設定

		spriteRenderer_ = new SpriteRenderer(D3DCompile_, device_);


		//--------------------------------------------------------------
		// メッセージリスナの設定

		Application.instance.addWindowMessageListener(this);
	}

	/++
	 + デストラクタ
	 +/
	~this()
	{
		Application.instance.removeWindowMessageListener(this);

		release();
	}

	/+
	 + 手動解放処理
	 +
	 + COMオブジェクト等は手動解放が必要
	 +/
	void release()
	{
		if (spriteRenderer_ !is null)
		{
			delete spriteRenderer_;
			spriteRenderer_ = null;
		}

		if (textureRenderTargetView_ !is null)
		{
			textureRenderTargetView_.Release();
			textureRenderTargetView_ = null;
		}

		if (textureResourceView_ !is null)
		{
			textureResourceView_.Release();
			textureResourceView_ = null;
		}

		if (texture_ !is null)
		{
			texture_.Release();
			texture_ = null;
		}

		if (backBufferRenerTargetView_ !is null)
		{
			backBufferRenerTargetView_.Release();
			backBufferRenerTargetView_ = null;
		}

		if (immediateContext_ !is null)
		{
			immediateContext_.Release();
			immediateContext_ = null;
		}

		if (device_ !is null)
		{
			device_.Release();
			device_ = null;
		}

		if (swapChain_ !is null)
		{
			swapChain_.Release();
			swapChain_ = null;
		}

		if (d3dcompiler_ !is null)
		{
			FreeLibrary(d3dcompiler_);
			d3dcompiler_ = null;
		}

		if (d3d11Dll_ !is null)
		{
			FreeLibrary(d3d11Dll_);
			d3d11Dll_ = null;
		}
	}
}

/++
 + スプライト描画ステータス
 +
 + Spriteを描画するためのDirect3D設定
 +/
class SpriteRenderer
{
private:

	// 定数バッファ
	static
	struct Constants
	{
		float x;           // Sprite.x
		float y;           // Sprite.y
		float centerX;     // Sprite.centerX
		float centerY;     // Sprite.centerY
		float scaleX;      // Sprite.scaleX
		float scaleY;      // Sprite.scaleY
		float inclineX;    // Sprite.inclineX
		float inclineY;    // Sprite.inclineY
		int u;             // Pattern.u
		int v;             // Pattern.v
		int width;         // Pattern.width
		int height;        // Pattern.height
		int textureWidth;  // Surface.width
		int textureHeight; // Surface.height
		int screenWidth;   // スクリーン幅の 1/2
		int screenHeight;  // スクリーン高さの 1/2
		uint rotation;     // Sprite.rotation
		uint color;        // Sprite.color
		float dummy0;      // 未使用(16byte境界に合わせないと定数バッファが作れない)
		float dummy1;      // 未使用(16byte境界に合わせないと定数バッファが作れない)
	}

	// 頂点フォーマット
	static
	struct Vertex
	{
		float x;
		float y;
		float z;
		float w;
		float u;
		float v;
	}

	// 頂点バッファ(システムメモリ)
	Vertex[4] vertices_;

	// 定数バッファ(システムメモリ)
	Constants constants_;

	// Input layout
	ID3D11InputLayout inputLayout_;

	// Vertex buffer
	ID3D11Buffer vertexBuffer_;

	// Vertex shader
	ID3D11VertexShader vertexShader_;

	// Vertex shader 定数バッファ
	ID3D11Buffer vertexShaderConstantsBuffer_;

	// Rasterizer設定
	ID3D11RasterizerState rasterizerState_;

	// Pixel shader
	ID3D11PixelShader pixelShader_;

	// Sampler
	ID3D11SamplerState samplerState_;

	// アルファブレンド設定
	ID3D11BlendState blendState_;

	/+
	 + コンストラクタ
	 +
	 + Params:
	 +  D3DCompile = 各種シェーダをコンパイルするための関数
	 +  device = ID3D11Device
	 +/
	this(D3DCompileFunction D3DCompile, ID3D11Device device)
	{
		ID3DBlob compiled;
		ID3DBlob errors;

		//--------------------------------------------------------------
		// Input assembler ステージで使用するデータを準備

		// Vertex shader をコンパイル
		string spriteVertexShaderSource = `
			cbuffer Constants : register(b0)
			{
				float2 position;
				float2 center;
				float2 scale;
				float2 incline;
				int4 pattern;
				int2 surface;
				int2 screen;
				int rotation;
				uint color;
			};

			struct VS_INPUT
			{
				float4 pos : POSITION;
				float2 uv : TEXCOORD;
			};

			struct VS_OUTPUT
			{
				float4 pos : SV_POSITION;
				float4 col : COLOR;
				float2 uv : TEXCOORD;
			};

			VS_OUTPUT spriteVertexShader(VS_INPUT input)
			{
				// 0 ～ 1の座標を0 ～ widthと0 ～ heightに伸ばす
				input.pos.xy *= pattern.zw;

				// 中心座標を移動する
				input.pos.xy -= center;

				// sinとcosを取得
				float s;
				float c;
				sincos(((float)(rotation & 0x1ff)) * 3.14159265f / 256.0f, s, c);

				// 初期値としてせん断 * 回転行列を作成する
				float2x2 mat =
				{
					c - (incline.y * s), s + (incline.y * c),
					(incline.x * c) - s, c + (incline.x * s),
				};

				// 拡大/縮小行列をかける
				mat._11_12 *= scale.x;
				mat._21_22 *= scale.y;

				// 作成した行列を座標にかける
				input.pos.xy = mul(input.pos.xy, mat);


				VS_OUTPUT output;

				// ローカル座標 → スクリーン座標変換
				input.pos.xy += position;
				input.pos.xy -= screen;
				input.pos.xy /= screen;
				input.pos.y = -input.pos.y;
				output.pos = input.pos;

				// テクスチャのUV座標計算
				input.uv *= pattern.zw;
				input.uv += pattern.xy;
				output.uv = input.uv / surface;

				// 色をfloat4に変換
				output.col.x = (color >> 16) & 0xff;
				output.col.y = (color >> 8) & 0xff;
				output.col.z = color & 0xff;
				output.col.w = (color >> 24) & 0xff;
				output.col /= 255;

				return output;
			}
		`;
		if (FAILED(D3DCompile(spriteVertexShaderSource.ptr, spriteVertexShaderSource.length, null, null, null, "spriteVertexShader\0".ptr, "vs_4_0\0".ptr, 0, 0, &compiled, &errors)))
		{
			string message = "";
			if (errors !is null)
			{
				message = (cast(char*)errors.GetBufferPointer())[0 .. errors.GetBufferSize()].idup;
				errors.Release();
			}
			throw new GameError("Failed compile vertex shader: " ~ message);
		}

		// 入力レイアウトを作成
		D3D11_INPUT_ELEMENT_DESC[2] inputElementDesc;
		inputElementDesc[0] = D3D11_INPUT_ELEMENT_DESC("POSITION\0".ptr, 0, DXGI_FORMAT.DXGI_FORMAT_R32G32B32A32_FLOAT, 0, 0, D3D11_INPUT_CLASSIFICATION.D3D11_INPUT_PER_VERTEX_DATA, 0);
		inputElementDesc[1] = D3D11_INPUT_ELEMENT_DESC("TEXCOORD\0".ptr, 0, DXGI_FORMAT.DXGI_FORMAT_R32G32_FLOAT, 0, 16, D3D11_INPUT_CLASSIFICATION.D3D11_INPUT_PER_VERTEX_DATA, 0);
		if (FAILED(device.CreateInputLayout(inputElementDesc.ptr, inputElementDesc.length, compiled.GetBufferPointer(), compiled.GetBufferSize(), &inputLayout_)))
		{
			compiled.Release();
			if (errors !is null)
				errors.Release();

			release();
			throw new GameError("Failed create input layout");
		}

		//--------------------------------------------------------------
		// Vertex Shader ステージで使用するデータを準備

		// Vertex shaderを作成
		if (FAILED(device.CreateVertexShader(compiled.GetBufferPointer(), compiled.GetBufferSize(), null, &vertexShader_)))
		{
			compiled.Release();
			if (errors !is null)
				errors.Release();
			release();
			throw new GameError("Failed create vertex shader");
		}

		compiled.Release();
		if (errors !is null)
			errors.Release();

		// 頂点データ
		vertices_[0] = Vertex(0, 0, 0, 1, 0, 0);
		vertices_[1] = Vertex(1, 0, 0, 1, 1, 0);
		vertices_[2] = Vertex(0, 1, 0, 1, 0, 1);
		vertices_[3] = Vertex(1, 1, 0, 1, 1, 1);

		D3D11_BUFFER_DESC vertexBufferDesc;
		vertexBufferDesc.ByteWidth = Vertex.sizeof * vertices_.length;
		vertexBufferDesc.Usage = D3D11_USAGE.D3D11_USAGE_DEFAULT;
		vertexBufferDesc.BindFlags = D3D11_BIND_FLAG.D3D11_BIND_VERTEX_BUFFER;
		vertexBufferDesc.CPUAccessFlags = 0;
		vertexBufferDesc.MiscFlags = 0;

		D3D11_SUBRESOURCE_DATA vertexBufferSubresource;
		vertexBufferSubresource.pSysMem = vertices_.ptr;
		if (FAILED(device.CreateBuffer(&vertexBufferDesc, &vertexBufferSubresource, &vertexBuffer_)))
		{
			release();
			throw new GameError("Failed create vertex buffer");
		}

		// 頂点シェーダ定数バッファを作成
		D3D11_BUFFER_DESC vertexShaderConstantsBufferDesc = D3D11_BUFFER_DESC(Constants.sizeof, D3D11_USAGE.D3D11_USAGE_DEFAULT, D3D11_BIND_FLAG.D3D11_BIND_CONSTANT_BUFFER, 0, 0);
		if (FAILED(device.CreateBuffer(&vertexShaderConstantsBufferDesc, null, &vertexShaderConstantsBuffer_)))
		{
			release();
			throw new GameError("Failed create constants buffer");
		}

		//--------------------------------------------------------------
		// Rasterizer ステージで使用するデータを準備


		// Rasterizer state
		D3D11_RASTERIZER_DESC rasterizerDesc;
		rasterizerDesc.FillMode = D3D11_FILL_MODE.D3D11_FILL_SOLID;
		rasterizerDesc.CullMode = D3D11_CULL_MODE.D3D11_CULL_NONE;
		rasterizerDesc.FrontCounterClockwise = FALSE;
		rasterizerDesc.DepthBias = 0;
		rasterizerDesc.DepthBiasClamp = 0;
		rasterizerDesc.SlopeScaledDepthBias = 0;
		rasterizerDesc.DepthClipEnable = FALSE;
		rasterizerDesc.ScissorEnable = FALSE;
		rasterizerDesc.MultisampleEnable = FALSE;
		rasterizerDesc.AntialiasedLineEnable = TRUE;
		if (FAILED(device.CreateRasterizerState(&rasterizerDesc, &rasterizerState_)))
		{
			release();
			throw new GameError("Failed create rasterizer state");
		}

		//--------------------------------------------------------------
		// Pixel Shader ステージで使用するデータを準備

		// Pixel shader
		string spritePixelShaderSource =
		`
			struct PS_INPUT
			{
				float4 pos : SV_POSITION;
				float4 col : COLOR;
				float2 uv : TEXCOORD;
			};

			Texture2D tex : register(t0);

			SamplerState sam : register(s0);

			float4 spritePixelShader(PS_INPUT input) : SV_Target
			{
				return tex.Sample(sam, input.uv) * input.col;
			}
		`;
		if (FAILED(D3DCompile(spritePixelShaderSource.ptr, spritePixelShaderSource.length, null, null, null, "spritePixelShader\0".ptr, "ps_4_0\0".ptr, 0, 0, &compiled, &errors)))
		{
			string message = "";
			if (errors !is null)
			{
				message = (cast(char*)errors.GetBufferPointer())[0 .. errors.GetBufferSize()].idup;
				errors.Release();
			}

			throw new GameError("Failed compile pixel shader: " ~ message);
		}

		// Pixel shaderを作成
		if (FAILED(device.CreatePixelShader(compiled.GetBufferPointer(), compiled.GetBufferSize(), null, &pixelShader_)))
		{
			compiled.Release();
			if (errors !is null)
				errors.Release();
			release();
			throw new GameError("Failed create vertex shader");
		}

		compiled.Release();
		if (errors !is null)
			errors.Release();


		D3D11_SAMPLER_DESC samplerDesc;
		samplerDesc.Filter = D3D11_FILTER.D3D11_FILTER_ANISOTROPIC;
		samplerDesc.AddressU = D3D11_TEXTURE_ADDRESS_MODE.D3D11_TEXTURE_ADDRESS_CLAMP;
		samplerDesc.AddressV = D3D11_TEXTURE_ADDRESS_MODE.D3D11_TEXTURE_ADDRESS_CLAMP;
		samplerDesc.AddressW = D3D11_TEXTURE_ADDRESS_MODE.D3D11_TEXTURE_ADDRESS_CLAMP;
		samplerDesc.MipLODBias = 0.0f;
		samplerDesc.MaxAnisotropy = 8;
		samplerDesc.ComparisonFunc = D3D11_COMPARISON_FUNC.D3D11_COMPARISON_ALWAYS;
		samplerDesc.BorderColor[0] = 0.0f;
		samplerDesc.BorderColor[1] = 0.0f;
		samplerDesc.BorderColor[2] = 0.0f;
		samplerDesc.BorderColor[3] = 0.0f;
		samplerDesc.MinLOD = 0.0f;
		samplerDesc.MaxLOD = D3D11_FLOAT32_MAX;
		if (FAILED(device.CreateSamplerState(&samplerDesc, &samplerState_)))
		{
			release();
		    throw new GameError("Failed create clamp sampler state");
		}

		//--------------------------------------------------------------
		// Output Merger ステージで使用するデータを準備

		// アルファブレンドの設定
		D3D11_BLEND_DESC blendStateDesc;
		blendStateDesc.AlphaToCoverageEnable = FALSE;
		blendStateDesc.IndependentBlendEnable = FALSE;
		blendStateDesc.RenderTarget[0].BlendEnable = TRUE;
		blendStateDesc.RenderTarget[0].SrcBlend = D3D11_BLEND.D3D11_BLEND_SRC_ALPHA;
		blendStateDesc.RenderTarget[0].DestBlend = D3D11_BLEND.D3D11_BLEND_INV_SRC_ALPHA;
		blendStateDesc.RenderTarget[0].BlendOp = D3D11_BLEND_OP.D3D11_BLEND_OP_ADD;
		blendStateDesc.RenderTarget[0].SrcBlendAlpha = D3D11_BLEND.D3D11_BLEND_ZERO;
		blendStateDesc.RenderTarget[0].DestBlendAlpha = D3D11_BLEND.D3D11_BLEND_ONE;
		blendStateDesc.RenderTarget[0].BlendOpAlpha = D3D11_BLEND_OP.D3D11_BLEND_OP_ADD;
		blendStateDesc.RenderTarget[0].RenderTargetWriteMask = D3D11_COLOR_WRITE_ENABLE.D3D11_COLOR_WRITE_ENABLE_ALL;
		if (FAILED(device.CreateBlendState(&blendStateDesc, &blendState_)))
		{
			release();
			throw new GameError("Failed create blend state");
		}
	}

	/+
	 + デストラクタ
	 +/
	~this()
	{
		release();
	}

	/+
	 + 描画対象のサイズを設定
	 +/
	void setScreenSize(int width, int height)
	{
		constants_.screenWidth = width / 2;
		constants_.screenHeight = height /2;
	}

	void render(ID3D11DeviceContext context, ID3D11ShaderResourceView textureResourceView)
	{
		// Input Assember ステージに入力するデータ形式を設定
		context.IASetInputLayout(inputLayout_);

		// Input Assember ステージに入力されたデータの使用方法を設定
		context.IASetPrimitiveTopology(D3D11_PRIMITIVE_TOPOLOGY.D3D11_PRIMITIVE_TOPOLOGY_TRIANGLESTRIP);

		// Input Assembler ステージに頂点をセット
		uint stride = Vertex.sizeof;
		uint offset = 0;

		// 頂点バッファをセット
		context.IASetVertexBuffers(0, 1, &vertexBuffer_, &stride, &offset);

		// Vertex Shader ステージに Vertex Shader を設定
		context.VSSetShader(vertexShader_, null, 0);

		// 定数バッファの内容を更新
		context.UpdateSubresource(vertexShaderConstantsBuffer_, 0, null, &constants_, 0, 0);

		// Vertex Shaderに定数バッファをセット
		context.VSSetConstantBuffers(0, 1, &vertexShaderConstantsBuffer_);

		// Rasterizer ステージの設定
		context.RSSetState(rasterizerState_);

		// Pixel Shaderに Pixel Shader を設定
		context.PSSetShader(pixelShader_, null, 0);

		// テクスチャをセット
		context.PSSetShaderResources(0, 1, &textureResourceView);

		// Pixel Shaderに Sampler を設定
		context.PSSetSamplers(0, 1, &samplerState_);

		// Output Merger ステージでアルファブレンドを設定
		context.OMSetBlendState(blendState_, null, 0xffffffff);


		// パイプラインを実行して描画する
		context.Draw(vertices_.length, 0);
	}

	void release()
	{
		if (blendState_ !is null)
		{
			blendState_.Release();
			blendState_ = null;
		}

		if (samplerState_ !is null)
		{
			samplerState_.Release();
			samplerState_ = null;
		}

		if (pixelShader_ !is null)
		{
			pixelShader_.Release();
			pixelShader_ = null;
		}

		if (rasterizerState_ !is null)
		{
			rasterizerState_.Release();
			rasterizerState_ = null;
		}

		if (vertexShaderConstantsBuffer_ !is null)
		{
			vertexShaderConstantsBuffer_.Release();
			vertexShaderConstantsBuffer_ = null;
		}

		if (vertexShader_ !is null)
		{
			vertexShader_.Release();
			vertexShader_ = null;
		}

		if (vertexBuffer_ !is null)
		{
			vertexBuffer_.Release();
			vertexBuffer_ = null;
		}

		if (inputLayout_ !is null)
		{
			inputLayout_.Release();
			inputLayout_ = null;
		}
	}
}

/++
 + ピクセル画像データのインターフェース
 +/
interface SurfaceSource
{
	@property
	{
		/// 画像の幅
		uint width();

		/// 画像の高さ
		uint height();
	}

	/++
	 + オープン
	 +
	 + ファイル等、読み出しに事前準備が必要な場合の処理
	 +
	 + Returns:
	 +  成功した場合にtrueを返す
	 +/
	bool open();

	/++
	 + クローズ
	 +
	 + openの後始末が必要な場合の処理
	 +/
	void close();

	/++
	 + データをメモリにロード
	 +
	 + Params:
	 +  buffer = 展開先メモリのポインタ
	 +  width = 展開先メモリ領域の幅
	 +  height = 展開先メモリ領域の高さ
	 +/
	void read(uint* buffer, uint width, uint height);
}

/++
 + ピクセル画像データ
 +
 + 実体はテクスチャ
 + テクスチャの縦横は2の累乗で作らなければならないため、
 + width, heightは指定したサイズより大きくなる場合がある
 +/
class Surface
{
	/++
	 + コンストラクタ
	 +/
	this()
	{
		// Graphicsの初期化を確認
		if (Graphics.instance is null)
			throw new GameError("Graphics is not initialized");
	}

	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  source = データソース
	 +/
	this(SurfaceSource source)
	{
		load(source);
	}

	/++
	 + デストラクタ
	 +/
	~this()
	{
		if (textureResourceView_ !is null)
		{
			textureResourceView_.Release();
			textureResourceView_ = null;
		}

		if (texture_ !is null)
		{
			texture_.Release();
			texture_ = null;
		}

		width_ = 0;
		height_ = 0;
		usedWidth_ = 0;
		usedHeight_ = 0;
	}

	@property
	{
		/// 幅
		const
		int width()
		{
			return width_;
		}

		/// 高さ
		const
		int height()
		{
			return height_;
		}

		/// 幅
		const
		int usedWidth()
		{
			return usedWidth_;
		}

		/// 高さ
		const
		int usedHeight()
		{
			return usedHeight;
		}

		ID3D11Texture2D texture()
		{
			return texture_;
		}

		ID3D11ShaderResourceView textureResourceView()
		{
			return textureResourceView_;
		}
	}

	/++
	 + ピクセルデータをロードする
	 +
	 + Params:
	 +  source = データソース
	 +/
	void load(SurfaceSource source)
	{
		// データソースをオープン
		if (source is null)
			return;

		if (!source.open())
			return;
		scope(exit) source.close();


		// 既存データをクリア
		if (textureResourceView_ !is null)
		{
			textureResourceView_.Release();
			textureResourceView_ = null;
		}

		if (texture_ !is null)
		{
			texture_.Release();
			texture_ = null;
		}

		width_ = 0;
		height_ = 0;
		usedWidth_ = 0;
		usedHeight_ = 0;

		// テクスチャのサイズ設定
		int width = source.width;
		int height = source.height;

		int size;

		size = 1;
		for (;;)
		{
			if (size >= width)
			{
				width = size;
				break;
			}
			size *= 2;
		}

		size = 1;
		for (;;)
		{
			if (size >= height)
			{
				height = size;
				break;
			}
			size *= 2;
		}


		// データを作業領域にロード
		if (width * height > workAreaLength_)
		{
			workAreaLength_ = width * height;
			workArea_ = cast(uint*)std.c.stdlib.realloc(workArea_, workAreaLength_ * 4);
		}

		source.read(workArea_, width, height);

		// 作成するテクスチャの設定
		D3D11_TEXTURE2D_DESC textureDesc;
		textureDesc.Width = width;
		textureDesc.Height = height;
		textureDesc.MipLevels = 1;
		textureDesc.ArraySize = 1;
		textureDesc.Format = DXGI_FORMAT.DXGI_FORMAT_R8G8B8A8_UNORM;
		textureDesc.SampleDesc.Count = 1;
		textureDesc.SampleDesc.Quality = 0;
		textureDesc.Usage = D3D11_USAGE.D3D11_USAGE_IMMUTABLE;
		textureDesc.BindFlags = D3D11_BIND_FLAG.D3D11_BIND_SHADER_RESOURCE;
		textureDesc.CPUAccessFlags = 0;

		// 初期化データとしてピクセルデータを設定
		D3D11_SUBRESOURCE_DATA textureSubsourceData = D3D11_SUBRESOURCE_DATA(cast(const(void)*)workArea_, width * 4);

		ID3D11Device device = Graphics.instance.device_;

		// 2Dテクスチャを作成
		if (FAILED(device.CreateTexture2D(&textureDesc, &textureSubsourceData, &texture_)))
			return;

		// 2Dテクスチャをシェーダリソースビューに関連づけ
		if (FAILED(device.CreateShaderResourceView(texture_, null, &textureResourceView_)))
		{
			texture_.Release();
			texture_ = null;
		}

		width_ = width;
		height_ = height;

		usedWidth_ = source.width;
		usedHeight_ = source.height;
	}

protected:

	// データ読み出し用のワークエリア
	static
	uint* workArea_;

	// ワークエリアのサイズ
	static
	uint workAreaLength_;

	// テクスチャ幅
	int width_;

	// テクスチャ高さ
	int height_;

	// テクスチャ内利用範囲の幅
	int usedWidth_;

	// テクスチャ内利用範囲の高さ
	int usedHeight_;

	// テクスチャインターフェース
	ID3D11Texture2D texture_;

	// テクスチャリソースビュー
	ID3D11ShaderResourceView textureResourceView_;

	/+
	 + 静的コンストラクタ
	 +/
	static
	this()
	{
		workAreaLength_ = 2048 * 2048;
		workArea_ = cast(uint*)std.c.stdlib.malloc(workAreaLength_ * 4);
	}

	/+
	 + 静的デストラクタ
	 +/
	static
	~this()
	{
		if (workArea_ !is null)
		{
			std.c.stdlib.free(workArea_);
			workArea_ = null;
		}
	}
}

/++
 + Surface内の矩形範囲
 +/
struct Pattern
{
	/// 左上点のサーフェス上でのX座標
	int u;

	/// 左上点のサーフェス上でのY座標
	int v;

	/// 幅
	int width;

	/// 高さ
	int height;

	/// サーフェス
	Surface surface;
}

/++
 + Patternの描画
 +/
struct Sprite
{
	enum DisplayType
	{
		NONE,      // 非表示
		NORMAL,    // 通常表示
		ADD_BLEND, // 加算合成
	}

	/// X軸表示座標
	float x = 0;

	/// Y軸表示座標
	float y = 0;

	/// X軸中心座標 この座標が表示座標で指定した点に来る 拡大・縮小・回転の中心にもなる
	float centerX = 0;

	/// Y軸中心座標 この座標が表示座標で指定した点に来る 拡大・縮小・回転の中心にもなる
	float centerY = 0;

	/// X軸せん断率
	float inclineX = 0;

	/// Y軸せん断率
	float inclineY = 0;

	/// X軸拡大率
	float scaleX = 1;

	/// Y軸拡大率
	float scaleY = 1;

	/// 回転角
	uint rotation = 0;

	/// 色
	uint color = 0xffffffff;

	// 表示形式
	DisplayType type = DisplayType.NORMAL;

	/// 表示する矩形範囲
	Pattern* pattern = null;

	/++
	 + 描画
	 +/
	void draw()
	{
		if (pattern is null || pattern.surface is null || pattern.surface.textureResourceView_ is null)
			return;

		// Spriteレンダリングオブジェクトを取得
		SpriteRenderer renderer = Graphics.instance.spriteRenderer_;

		// Spriteレンダリング設定
		SpriteRenderer.Constants* constants = &renderer.constants_;
		constants.x = x;
		constants.y = y;
		constants.centerX = centerX;
		constants.centerY = centerY;
		constants.scaleX = scaleX;
		constants.scaleY = scaleY;
		constants.inclineX = inclineX;
		constants.inclineY = inclineY;
		constants.u = pattern.u;
		constants.v = pattern.v;
		constants.width = pattern.width;
		constants.height = pattern.height;
		constants.textureWidth = pattern.surface.width;
		constants.textureHeight = pattern.surface.height;
		constants.rotation = rotation;
		constants.color = color;

		ID3D11DeviceContext context = Graphics.instance.immediateContext_;

		// 描画処理
		renderer.render(context, pattern.surface.textureResourceView_);
	}
}

/++
 + 独自形式
 +/
class CompressedGraphicsFile : SurfaceSource
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  filePath = ファイルパス
	 +  offset = ファイル先頭からのオフセット。1ファイルに複数のファイルをまとめた時に使用
	 +/
	this(string filePath, uint offset)
	{
		filePath_ = filePath ~ 0;
		fileOffset_ = offset;
	}

	/++
	 + デストラクタ
	 +/
	~this()
	{
		close();
	}

	@property
	{
		uint width()
		{
			return width_;
		}

		uint height()
		{
			return height_;
		}

		string filePath()
		{
			return filePath_[0 .. $ - 1];
		}
	}

	bool open()
	{
		// ファイル自体をオープン
		file_ = std.c.stdio.fopen(filePath_.ptr, "rb\0".ptr);
		if (file_ is null)
			return false;

		// アーカイブ内のファイル先頭位置へ移動
		if (fileOffset_ > 0)
			std.c.stdio.fseek(file_, fileOffset_, std.c.stdio.SEEK_SET);

		// 縦横サイズを取得
		ushort[2] data;
		std.c.stdio.fread(data.ptr, ushort.sizeof, data.length, file_);

		width_ = data[0];
		height_ = data[1];

		return true;
	}

	void close()
	{
		std.c.stdio.fclose(file_);

		width_ = 0;
		height_ = 0;
	}

	void read(uint* buffer, uint width, uint height)
	{
		// 一致する仕様
		if (width_ != width || height_ != height)
			return;

		// 圧縮済みサイズを取得
		uint compressedLength;
		std.c.stdio.fread(&compressedLength, uint.sizeof, 1, file_);

		// データ読み出しバッファを確保
		ubyte* compressed = cast(ubyte*)std.c.stdlib.malloc(compressedLength);
		scope(exit) std.c.stdlib.free(compressed);

		// データを読み出し
		std.c.stdio.fread(compressed, ubyte.sizeof, compressedLength, file_);

		// 読み込みカーソル
		ubyte* source = compressed;

		// テーブル読み出し
		uint[256] rangeTable;
		if (source[0] == 4)
		{
			source += 4;

			for (uint i = 0; i < 256; ++i)
				rangeTable[i] = source[i * 4] | (source[i * 4 + 1] << 8) | (source[i * 4 + 2] << 16) | (source[i * 4 + 3] << 24);

			source += 256 * 4;
		}
		else if (source[0] == 3)
		{
			source += 4;

			for (uint i = 0; i < 256; ++i)
				rangeTable[i] = source[i * 3] | (source[i * 3 + 1] << 8) | (source[i * 3 + 2] << 16);

			source += 256 * 3;
		}
		else if (source[0] == 2)
		{
			source += 4;

			for (uint i = 0; i < 256; ++i)
				rangeTable[i] = source[i * 2] | (source[i * 2 + 1] << 8);

			source += 256 * 2;
		}
		else
		{
			source += 4;

			for (uint i = 0; i < 256; ++i)
				rangeTable[i] = source[i];

			source += 256;
		}

		// 下限と解凍後のサイズをセット
		uint[256] lowTable;
		uint outputLength;
		for (uint i; i < 256; ++i)
		{
			lowTable[i] = outputLength;
			outputLength += rangeTable[i];
		}

		// 最初に0が入っているので読み飛ばす
		++source;

		// lowを初期化
		uint low = (source[0] << 24) | (source[1] << 16) | (source[2] << 8) | source[3];
		source += 4;

		// rangeを初期化
		uint range = 0xffffffff;

		ubyte* destination = cast(ubyte*)buffer;
		ubyte* end = destination + width * height * 4;
		while (destination < end)
		{
			uint flag = 0x100;
			{
				uint tmp = range / outputLength;

				uint code = low / tmp;

				// 値が入る範囲を検索
				int charactor = 0;
				int jump = 256;
				for (;;)
				{
					if (lowTable[charactor] > code)
					{
						jump >>= 1;
						charactor -= jump;
					}
					else if (lowTable[charactor] + rangeTable[charactor] <= code)
					{
						jump >>= 1;
						charactor += jump;
					}
					else
						break;
				}

				// 出力
				flag |= cast(ubyte)charactor;

				// Rangeを計算
				range = rangeTable[charactor] * tmp;

				//	Lowを計算
				low -= lowTable[charactor] * tmp;

				//	Rangeが小さくなったら256倍する
				while ((range & 0xff000000) == 0)
				{
					range <<= 8;
					low <<= 8;
					low |= *source++;
				}
			}

			while ((flag & 0x10000) == 0)
			{
				if ((flag & 0x80) == 0)
				{
					if (destination >= end)
						return;

					uint tmp = range / outputLength;

					uint code = low / tmp;

					// 値が入る範囲を検索
					int charactor = 0;
					int jump = 256;
					for (;;)
					{
						if (lowTable[charactor] > code)
						{
							jump >>= 1;
							charactor -= jump;
						}
						else if (lowTable[charactor] + rangeTable[charactor] <= code)
						{
							jump >>= 1;
							charactor += jump;
						}
						else
							break;
					}

					// 出力
					*destination++ = cast(ubyte)charactor;

					// Rangeを計算
					range = rangeTable[charactor] * tmp;

					//	Lowを計算
					low -= lowTable[charactor] * tmp;

					//	Rangeが小さくなったら256倍する
					while ((range & 0xff000000) == 0)
					{
						range <<= 8;
						low <<= 8;
						low |= *source++;
					}
				}
				else
				{
					ubyte[2] data;
					for (uint i; i < 2; ++i)
					{
						uint tmp = range / outputLength;

						uint code = low / tmp;

						// 値が入る範囲を検索
						int charactor = 0;
						int jump = 256;
						for (;;)
						{
							if (lowTable[charactor] > code)
							{
								jump >>= 1;
								charactor -= jump;
							}
							else if (lowTable[charactor] + rangeTable[charactor] <= code)
							{
								jump >>= 1;
								charactor += jump;
							}
							else
								break;
						}

						// 出力
						data[i] = cast(ubyte)charactor;

						// Rangeを計算
						range = rangeTable[charactor] * tmp;

						//	Lowを計算
						low -= lowTable[charactor] * tmp;

						//	Rangeが小さくなったら256倍する
						while ((range & 0xff000000) == 0)
						{
							range <<= 8;
							low <<= 8;
							low |= *source++;
						}
					}

					uint length = (data[0] & 15) + 4;
					uint distance = (((data[0] << 4) & 0xff00) | data[1]) + 4;

					const(ubyte)* s = destination - distance;
					for (uint i; i < length; ++i)
						*destination++ = s[i];
				}

				flag <<= 1;
			}
		}
	}

private:

	// 幅
	uint width_;

	// 高さ
	uint height_;

	// アーカイブ中での個別ファイルの位置
	uint fileOffset_;

	FILE* file_;

	// ファイルパス
	string filePath_;
}

/++
 + フォントファイル
 +/
class FontFile : SurfaceSource
{
	/// メッセージ
	string message;

	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  filePath = ファイルパス
	 +  offset = ファイル先頭からのオフセット。1ファイルに複数のファイルをまとめた時に使用
	 +/
	this(string filePath, uint offset)
	{
		uint length;

		// ファイルをオープン
		auto file = new BufferedFile(filePath, FileMode.In);
		scope(exit) file.close;

		// 読み込み位置まで移動
		file.seek(offset, SeekPos.Current);

		// 1文字辺りのサイズをロード
		file.read(charactorWidth_);
		file.read(charactorHeight_);

		// 文字数をロード
		file.read(length);

		// 文字コードをロード
		ushort[] charactors = new ushort[length];
		file.read(cast(ubyte[])charactors);

		// 文字データのアドレスをロード
		uint[] addresses = new uint[length];
		file.read(cast(ubyte[])addresses);

		// 文字コードからアドレスへの変換テーブルを作成
		for (uint i; i < length; ++i)
			addresses_[charactors[i]] = addresses[i];

		// Huffman木のノード数をロード
		file.read(length);

		// Huffman木をロード
		huffmanTree_ = new HuffmanNode[length];
		file.read(cast(ubyte[])huffmanTree_);

		// 圧縮されたピクセルデータのサイズをロード
		file.read(length);

		// 圧縮されたピクセルデータをロード
		data_ = new ubyte[length];
		file.read(data_);
	}

	bool open()
	{
		return true;
	}

	void close()
	{
	}

	void read(uint* buffer, uint width, uint height)
	{
		// 非効率だがロジックが単純になるように一旦全クリアする
		for (uint y; y < height_; ++y)
			for (uint x; x < width_; ++x)
				buffer[y * width + x] = 0;

		uint originX;
		uint originY;

		// Huffman木のrootを取得
		ushort root = cast(ushort)(huffmanTree_.length - 1);

		for (uint i; i < message.length; ++i)
		{
			ushort charactor = message[i];

			// 改行コードの判定
			if (charactor == 0x0A)
			{
				originX = 0;
				originY += charactorHeight_;
				continue;
			}

			// 全角半角の判定
			uint charactorWidth;
			if (charactor < 0x80)
				charactorWidth = charactorWidth_ / 2;
			else
			{
				charactorWidth = charactorWidth_;

				charactor <<= 8;
				charactor |= message[++i];
			}

			// 横がいっぱいならば次の行へ
			if (originX + charactorWidth > width_)
			{
				originX = 0;
				originY += charactorHeight_;
			}

			// 縦がいっぱいなら終わり
			if (originY + charactorHeight_ > height_)
				return;

			// 圧縮データの先頭を取得
			ubyte* source = &data_[addresses_[charactor]];

			// 次に読むbit位置
			ubyte position = 7;

			// bitデータ
			ubyte bits = *source++;

			uint* write = &buffer[originY * width + originX];
			for (uint y; y < charactorHeight_; ++y)
			{
				for (uint x; x < charactorWidth; ++x)
				{
					// カーソルをHuffman木のrootにセット
					ushort current = root;

					// 葉に達するまで木をたどる
					for (;;)
					{
						// bitの0 or 1で左 or 右
						if (((bits >> position) & 1) == 0)
							current = huffmanTree_[current].left;
						else
							current = huffmanTree_[current].right;

						// 8bit読み出し終わったら1Byte補充する
						if (position == 0)
						{
							position = 7;
							bits = *source++;
						}
						else
							--position;

						// 葉に達したら文字を取得
						if (huffmanTree_[current].left == 0xffff)
						{
							write[x] = (huffmanTree_[current].data << 24) | 0x00ffffff;
							break;
						}
					}
				}
				write += width;
			}

			originX += charactorWidth;
		}
	}

	@property
	{
		uint width()
		{
			return width_;
		}

		void width(uint w)
		{
			width_ = w;
		}

		uint height()
		{
			return height_;
		}

		void height(uint h)
		{
			height_ = h;
		}

		uint charactorWidth()
		{
			return charactorWidth_;
		}

		uint charactorHeight()
		{
			return charactorHeight_;
		}
	}

	/++
	 + データ設定のショートカット
	 +
	 + Params:
	 +  width = 幅
	 +  height = 高さ
	 +  msg = メッセージ
	 +/
	void set(uint width, uint height, string msg)
	{
		width_ = width;
		height_ = height;
		message = msg;
	}

	/++
	 + メッセージをセットした時に各行の横幅がいくつになるか計算
	 +
	 + Params:
	 +  message = メッセージ
	 +
	 + Returns:
	 +  計算結果
	 +/
	uint[] calculateLineWidth(string message)
	{
		uint[] output;

		uint line;
		for (uint i; i < message.length; ++i)
		{
			// 改行コードの判定
			if (message[i] == 0x0A)
			{
				if (output.length == line)
					output ~= 0;
				++line;
				continue;
			}

			// 全角半角の判定
			uint charactorWidth;
			if (message[i] < 0x80)
				charactorWidth = charactorWidth_ / 2;
			else
			{
				charactorWidth = charactorWidth_;
				++i;
			}

			if (charactorWidth > width_)
				return output;

			if (output.length == 0)
				output ~= charactorWidth;
			else
			{
				if (output[line] + charactorWidth > width_)
				{
					output ~= charactorWidth;
					++line;
				}
				else
					output[line] += charactorWidth;
			}
		}

		return output;
	}

	/++
	 + メッセージをセットした時に各行の横幅がいくつになるか計算
	 +
	 + Params:
	 +  message = メッセージ
	 +  output = 計算結果を入れる配列
	 +/
	void calculateLineWidth(string message, uint[] output)
	{
		if (output is null || output.length == 0)
			return;

		// 最初の行をクリア
		uint line;
		output[line] = 0;

		for (uint i; i < message.length; ++i)
		{
			// 改行コードの判定
			if (message[i] == 0x0A)
			{
				// 次の行へ
				if (++line >= output.length)
					return;
				output[line] = 0;

				continue;
			}

			// 全角半角の判定
			uint charactorWidth;
			if (message[i] < 0x80)
				charactorWidth = charactorWidth_ / 2;
			else
			{
				charactorWidth = charactorWidth_;
				++i;
			}

			while (output[line] + charactorWidth > width_)
			{
				// 次の行へ
				if (++line >= output.length)
					return;
				output[line] = 0;
			}

			output[line] += charactorWidth;
		}

		// 残った行を0クリア
		for (++line; line < output.length; ++line)
			output[line] = 0;
	}

private:

	// Huffuman木のノード
	static
	struct HuffmanNode
	{
		ushort left;
		ushort right;
		ubyte data;
		ubyte[3] dummy;
	}

	// 幅
	uint width_;

	// 高さ
	uint height_;

	// 1文字の幅
	uint charactorWidth_;

	// 1文字の高さ
	uint charactorHeight_;

	// 文字列解凍用Huffman木
	HuffmanNode[] huffmanTree_;

	// 文字コードからデータアドレスへの変換テーブル
	uint[65536] addresses_;

	// 圧縮されたピクセルデータ
	ubyte[] data_;
}
