/++
 + Vorbis codec header module for D programming language
 +
 + Use, distribution and reproduction of this library source is 
 + governed by a BSD-STYLE souce license included with this source in 'COPYING'.
 + Please read these terms before distributing.
 +
 + Copyright(C) 2009-2013 Takuya Kurosawa
 +
 + Original code:
 +  codec.h (libvorbis-1.3.3)
 +  THE OggVorbis SOURCE CODE IS (C) COPYRIGHT 1994-2001
 +  by the Xiph.Org Foundation http://www.xiph.org/
 +/

module org.xiph.libvorbis.codec;

import org.xiph.libogg.types;
import org.xiph.libogg.ogg;

struct vorbis_info
{
	int ver;
	int channels;
	int rate;

	/+
	 + The below bitrate declarations are *hints*.
	 + Combinations of the three values carry the following implications:
	 +
	 + all three set to the same value:
	 +     implies a fixed rate bitstream
	 + only nominal set:
	 +     implies a VBR stream that averages the nominal bitrate.  No hard
	 +     upper/lower limit
	 + upper and or lower set:
	 +     implies a VBR bitstream that obeys the bitrate limits. nominal
	 +     may also be set to give a nominal rate.
	 + none set:
	 +     the coder does not care to speculate.
	 +/
	int bitrate_upper;
	int bitrate_nominal;
	int bitrate_lower;
	int bitrate_window;

	void* codec_setup;
}


/++
 + vorbis_dsp_state buffers the current vorbis audio analysis/synthesis state.
 + The DSP state belongs to a specific logical bitstream
 +/
struct vorbis_dsp_state
{
  int analysisp;
  vorbis_info* vi;

  float** pcm;
  float** pcmret;
  int pcm_storage;
  int pcm_current;
  int pcm_returned;

  int preextrapolate;
  int eofflag;

  int lW;
  int W;
  int nW;
  int centerW;

  ogg_int64_t granulepos;
  ogg_int64_t sequence;

  ogg_int64_t glue_bits;
  ogg_int64_t time_bits;
  ogg_int64_t floor_bits;
  ogg_int64_t res_bits;

  void* backend_state;
}

// necessary stream state for linking to the framing abstraction
struct vorbis_block
{
	// this is a pointer into local storage
	float** pcm;

	oggpack_buffer opb;

	int lW;
	int W;
	int nW;
	int pcmend;
	int mode;

	int eofflag;
	ogg_int64_t granulepos;
	ogg_int64_t sequence;

	// For read-only access of configuration
	vorbis_dsp_state* vd;

	// local storage to avoid remallocing; it's up to the mapping to structure it
	void* localstore;
	int localtop;
	int localalloc;
	int totaluse;
	alloc_chain* reap;

	// bitmetrics for the frame
	int glue_bits;
	int time_bits;
	int floor_bits;
	int res_bits;

	void* internal;
}

/++
 + vorbis_block is a single block of data to be processed as part of
 + the analysis/synthesis stream; it belongs to a specific logical
 + bitstream, but is independant from other vorbis_blocks belonging to that logical bitstream.
 +/
struct alloc_chain
{
	void* ptr;
	alloc_chain* next;
}

/++
 + vorbis_info contains all the setup information specific to the
 + specific compression/decompression mode in 
 + progress (eg, psychoacoustic settings, channel setup, options, codebook etc).
 + vorbis_info and substructures are in backends.h.
 + the comments are not part of vorbis_info so that vorbis_info can be static storage
 +/
struct vorbis_comment
{
	// unlimited user comment fields.  libvorbis writes 'libvorbis' whatever vendor is set to in encode
	char** user_comments;
	int* comment_lengths;
	int comments;
	char* vendor;
}

/+
 + libvorbis encodes in two abstraction layers; first we perform DSP
 + and produce a packet (see docs/analysis.txt).  The packet is then
 + coded into a framed OggSquish bitstream by the second layer (see
 + docs/framing.txt).  Decode is the reverse process; we sync/frame
 + the bitstream and extract individual packets, then decode the
 +  packet back into PCM audio.
 +
 + The extra framing/packetizing is used in streaming formats, such as
 + files.  Over the net (such as with UDP), the framing and
 + packetization aren't necessary as they're provided by the transport
 + and the streaming layer is not used
 +/
extern (C)
{
	/+
	 + Vorbis PRIMITIVES: general
	 +/
	void vorbis_info_init(vorbis_info* vi);
	void vorbis_info_clear(vorbis_info* vi);
	int vorbis_info_blocksize(vorbis_info* vi, int zo);
	void vorbis_comment_init(vorbis_comment* vc);
	void vorbis_comment_add(vorbis_comment* vc, const(char)* comment);
	void vorbis_comment_add_tag(vorbis_comment* vc,  const(char)* tag, const(char)* contents);
	char* vorbis_comment_query(vorbis_comment* vc, const(char)* tag, int count);
	int vorbis_comment_query_count(vorbis_comment* vc, const(char)* tag);
	void vorbis_comment_clear(vorbis_comment* vc);

	int vorbis_block_init(vorbis_dsp_state* v, vorbis_block* vb);
	int vorbis_block_clear(vorbis_block* vb);
	void vorbis_dsp_clear(vorbis_dsp_state* v);
	double vorbis_granule_time(vorbis_dsp_state* v, ogg_int64_t granulepos);

	const(char)* vorbis_version_string();

	/+
	 + Vorbis PRIMITIVES: analysis/DSP layer
	 +/
	int vorbis_analysis_init(vorbis_dsp_state* v, vorbis_info* vi);
	int vorbis_commentheader_out(vorbis_comment* vc, ogg_packet* op);
	int vorbis_analysis_headerout(vorbis_dsp_state* v, vorbis_comment* vc, ogg_packet* op, ogg_packet* op_comm, ogg_packet* op_code);
	float** vorbis_analysis_buffer(vorbis_dsp_state* v, int vals);
	int vorbis_analysis_wrote(vorbis_dsp_state* v, int vals);
	int vorbis_analysis_blockout(vorbis_dsp_state* v, vorbis_block* vb);
	int vorbis_analysis(vorbis_block* vb, ogg_packet* op);
	int vorbis_bitrate_addblock(vorbis_block* vb);
	int vorbis_bitrate_flushpacket(vorbis_dsp_state* vd, ogg_packet* op);

	/+
	 + Vorbis PRIMITIVES: synthesis layer
	 +/
	int vorbis_synthesis_idheader(ogg_packet* op);
	int vorbis_synthesis_headerin(vorbis_info* vi, vorbis_comment* vc, ogg_packet* op);

	int vorbis_synthesis_init(vorbis_dsp_state* v, vorbis_info* vi);
	int vorbis_synthesis_restart(vorbis_dsp_state* v);
	int vorbis_synthesis(vorbis_block* vb, ogg_packet* op);
	int vorbis_synthesis_trackonly(vorbis_block* vb, ogg_packet* op);
	int vorbis_synthesis_blockin(vorbis_dsp_state* v, vorbis_block* vb);
	int vorbis_synthesis_pcmout(vorbis_dsp_state* v, float* **pcm);
	int vorbis_synthesis_lapout(vorbis_dsp_state* v, float* **pcm);
	int vorbis_synthesis_read(vorbis_dsp_state* v, int samples);
	int vorbis_packet_blocksize(vorbis_info* vi, ogg_packet* op);

	int vorbis_synthesis_halfrate(vorbis_info* v, int flag);
	int vorbis_synthesis_halfrate_p(vorbis_info* v);
}

/+
 + Vorbis ERRORS and return codes
 +/
immutable int OV_FALSE = -1;
immutable int OV_EOF   = -2;
immutable int OV_HOLE  = -3;

immutable int OV_EREAD      = -128;
immutable int OV_EFAULT     = -129;
immutable int OV_EIMPL      = -130;
immutable int OV_EINVAL     = -131;
immutable int OV_ENOTVORBIS = -132;
immutable int OV_EBADHEADER = -133;
immutable int OV_EVERSION   = -134;
immutable int OV_ENOTAUDIO  = -135;
immutable int OV_EBADPACKET = -136;
immutable int OV_EBADLINK   = -137;
immutable int OV_ENOSEEK    = -138;

